"""
   Common file for running validations.
"""
from datetime import datetime
import json
import re
from time import time
import pymongo
import copy 

from collections import defaultdict
from processor.logging.log_handler import getlogger
from processor.comparison.interpreter import Comparator
from processor.helper.json.json_utils import get_field_value, get_json_files,\
    json_from_file, TEST, collectiontypes, SNAPSHOT, JSONTEST, MASTERTEST, get_field_value_with_default
from processor.helper.config.config_utils import config_value, get_test_json_dir,\
    DATABASE, DBNAME, framework_dir, EXCLUSION
from processor.database.database import create_indexes, COLLECTION,\
    sort_field, get_documents
from processor.reporting.json_output import dump_output_results, update_output_testname
from processor.helper.config.rundata_utils import get_dbtests, get_from_currentdata
from processor.connector.populate_json import pull_json_data
from processor.connector.special_compliance.compliances import COMPLIANCES

logger = getlogger()

def get_snapshot_file(snapshot_file, container, dbname, filesystem):
    snapshot_json_data = {}
    if filesystem:
        file_name = '%s.json' % snapshot_file if snapshot_file and not \
            snapshot_file.endswith('.json') else snapshot_file
        snapshot_file = '%s/%s/%s' % (get_test_json_dir(), container, file_name)
        snapshot_json_data = json_from_file(snapshot_file)
    else:
        parts = snapshot_file.split('.')
        collection = config_value(DATABASE, collectiontypes[SNAPSHOT])
        qry = {'container': container, 'name': parts[0]}
        sort = [sort_field('timestamp', False)]
        docs = get_documents(collection, dbname=dbname, sort=sort, query=qry, limit=1)
        logger.info('Number of Snapshot Documents: %s', len(docs))
        if docs and len(docs):
            snapshot_json_data = docs[0]['json']
    return snapshot_json_data

def get_snapshot_id_to_collection_dict(snapshot_file, container, dbname, filesystem=True):
    snapshot_data = {}
    snapshot_json_data = get_snapshot_file(snapshot_file, container, dbname, filesystem)
    if snapshot_json_data and "connector" in snapshot_json_data and "remoteFile" in snapshot_json_data \
        and snapshot_json_data["connector"] and snapshot_json_data["remoteFile"]:
            _, pull_response = pull_json_data(snapshot_json_data)
            if not pull_response:
                return snapshot_data

    snapshots = get_field_value(snapshot_json_data, 'snapshots')
    if not snapshots:
        logger.info("\tERROR: Snapshot does not contain snapshots, continuing...")
        return snapshot_data
    for snapshot in snapshots:
        nodes = get_field_value(snapshot, 'nodes')
        if not nodes:
            logger.info("No nodes in snapshot, continuing to next!...")
            continue
        for node in nodes:
            sid = get_field_value(node, 'snapshotId')
            coll = node['collection'] if 'collection' in node else COLLECTION
            collection = coll.replace('.', '').lower()
            snapshot_data[sid] = collection
            if get_dbtests():
                create_indexes(collection, dbname, [('timestamp', pymongo.TEXT)])
    return snapshot_data


def run_validation_test(version, container, dbname, collection_data, testcase, excludedTestIds, includeTests):
    comparator = Comparator(version, container, dbname, collection_data, testcase, excludedTestIds, includeTests)
    results = comparator.validate()
    if isinstance(results, list):
        for result in results:
            result["result_id"] = "%s_%s" % (re.sub(r"(?=[-_\s\.]).", "", container).lower(), str(int(datetime.now().timestamp())))
            result.update(testcase)
        return results
    else:
        results.update(testcase)
        return[results]


def run_file_validation_tests(test_file, container, filesystem=True, snapshot_status=None):
    # logger.info("*" * 50)
    logger.info("\tTEST: %s", test_file)
    dirpath = None
    test_json_data = json_from_file(test_file)
    if not test_json_data:
        logger.info("\t\tTest file %s looks to be empty, next!...", test_file)

    if test_json_data and "connector" in test_json_data and "remoteFile" in test_json_data and test_json_data["connector"] and test_json_data["remoteFile"]:
        dirpath, pull_response = pull_json_data(test_json_data)
        if not pull_response:
            return {}

    # singletest = get_from_currentdata(SINGLETEST)
    # if singletest:
    #     testsets = get_field_value_with_default(test_json_data, 'testSet', [])
    #     for testset in testsets:
    #         newtestcases = []
    #         for testcase in testset['cases']:
    #             if ('testId' in testcase and testcase['testId'] == singletest) or \
    #                     ('masterTestId' in testcase and testcase['masterTestId'] == singletest):
    #                 newtestcases.append(testcase)
    #         testset['cases'] = newtestcases
    resultset = run_json_validation_tests(test_json_data, container, filesystem, snapshot_status, dirpath=dirpath)
    finalresult = True
    if resultset:
        snapshot = test_json_data['snapshot'] if 'snapshot' in test_json_data else ''
        # if singletest:
        #     print(json.dumps(resultset, indent=2))
        # else:
        #     dump_output_results(resultset, container, test_file, snapshot, filesystem)
        dump_output_results(resultset, container, test_file, snapshot, filesystem)
        for result in resultset:
            if 'result' in result:
                if not re.match(r'passed', result['result'], re.I):
                    finalresult = False
                    break
    else:
        # TODO: NO test cases in this file.
        # LOG HERE that no test cases are present in this file.
        finalresult = False
    return finalresult


def run_json_validation_tests(test_json_data, container, filesystem=True, snapshot_status=None, dirpath=None):
    resultset = []
    if not test_json_data:
        return resultset
    if not snapshot_status:
        snapshot_status = {}
    # logger.info("Valid Test JSON data")
    logger.debug(json.dumps(test_json_data, indent=2))
    testsets = get_field_value(test_json_data, 'testSet')
    if not testsets or not isinstance(testsets, list):
        logger.info("Test json does not contain testset, next!...")
        return resultset
    dbname = config_value(DATABASE, DBNAME)
    # Populate the snapshotId => collection for the snapshot.json in the test file.
    collection_data = get_snapshot_id_to_collection_dict(test_json_data['snapshot'],
                                                         container, dbname, filesystem)
    if test_json_data['snapshot'] in snapshot_status:
        current_snapshot_status = snapshot_status[test_json_data['snapshot']]
    else:
        current_snapshot_status = {}

    includeTestConfig = get_from_currentdata("INCLUDETESTS")
    includeTests = []
    if includeTestConfig:
        includeTests = get_from_currentdata("TESTIDS")
    excludedTestIds = {}
    exclusions = get_from_currentdata(EXCLUSION).get('exclusions', [])
    for exclusion in exclusions:
        if 'exclusionType' in exclusion and exclusion['exclusionType'] and exclusion['exclusionType'] == 'single':
            if 'masterTestID' in exclusion and exclusion['masterTestID'] \
                    and 'paths' in exclusion and exclusion['paths']:
                toAdd = False
                if includeTestConfig:
                    if exclusion['masterTestID'] not in includeTests:
                        toAdd = True
                else:
                    toAdd =  True
                if toAdd:
                    if exclusion['masterTestID'] not in excludedTestIds:
                        excludedTestIds[exclusion['masterTestID']] = exclusion['paths']
                    else:
                        excludedTestIds[exclusion['masterTestID']].extend(exclusion['paths'])

    skip = 0
    limit = 10
    dumpsize = 10
    for testset in testsets:
        version = get_field_value(testset, 'version')
        testcases = get_field_value(testset, 'cases')
        if not testcases or not isinstance(testcases, list):
            logger.info("No testcases in testSet!...")
            continue
        for testcase in testset['cases']:
            if "status" in testcase and testcase["status"] == "disable":
                continue
            else:
                testcase["status"] = "enable"
            if dirpath:
                testcase['dirpath'] = dirpath
            results = run_validation_test(version, container, dbname, collection_data,
                                             testcase, excludedTestIds, includeTests)
            resultset.extend(results)
            if not filesystem:
                if len(resultset) >= limit:
                    dump_output_results(resultset[skip:limit], container, test_file="", snapshot="", filesystem=False)
                    skip = limit
                    limit = skip + dumpsize
        
        if not filesystem and len(resultset) >= (skip+1):
            dump_output_results(resultset[skip:], container, test_file="", snapshot="", filesystem=False)
            skip = len(resultset)
            limit = skip + dumpsize
        
    return resultset


def run_container_validation_tests(container, dbsystem=True, snapshot_status=None):
    if not snapshot_status:
        snapshot_status = {}
    if dbsystem:
        return run_container_validation_tests_database(container, snapshot_status)
    else:
        return run_container_validation_tests_filesystem(container, snapshot_status)


def run_container_validation_tests_filesystem(container, snapshot_status=None):
    """Get test files from the filesystem."""
    # logger.info("Starting validation tests")
    logger.info("VALIDATION:")
    logger.info("\tCollection: %s,  Type: FILESYSTEM", container)
    reporting_path = config_value('REPORTING', 'reportOutputFolder')
    json_dir = '%s/%s/%s' % (framework_dir(), reporting_path, container)
    logger.info('\tLOCATION: %s', json_dir)
    test_files = get_json_files(json_dir, JSONTEST)
    # logger.info('\n'.join(test_files))
    result = True
    for test_file in test_files:
        logger.info('\tCOLLECTION: %s', test_file)
        val = run_file_validation_tests(test_file, container, True, snapshot_status)
        result = result and val
    if test_files:
        # return the result value if "test" file is processed collection
        logger.critical("VALIDATION COMPLETE:")
        return result

    # mastertest files
    test_files = get_json_files(json_dir, MASTERTEST)
    # logger.info('\n'.join(test_files))
    if not test_files:
        logger.error("ERROR: No `test` or `mastertest` file found. collection should contain either `test` or `mastertest` file")
        return False

    finalresult = result
    for test_file in test_files:
        logger.info('\tCOLLECTION: %s', test_file)
        # logger.info("*" * 50)
        # logger.info("validator tests: %s", test_file)
        dirpath = None
        test_json_data = json_from_file(test_file)
        if not test_json_data:
            logger.info("Test file %s looks to be empty, next!...", test_file)
            continue

        if "connector" in test_json_data and "remoteFile" in test_json_data and test_json_data["connector"] and test_json_data["remoteFile"]:
            dirpath, pull_response = pull_json_data(test_json_data)
            if not pull_response:
                return {}

        snapshot_key = '%s_gen' % test_json_data['masterSnapshot']
        mastersnapshots = defaultdict(list)
        snapshot_data = snapshot_status[snapshot_key] if snapshot_key in snapshot_status else {}
        for snapshot_id, mastersnapshot_id in snapshot_data.items():
            if isinstance(mastersnapshot_id, list):
                for master_snapshot_id in mastersnapshot_id:
                    mastersnapshots[master_snapshot_id].append(snapshot_id)
            elif isinstance(mastersnapshot_id, str):
                mastersnapshots[mastersnapshot_id].append(snapshot_id)
        if not mastersnapshots:
            logger.error("No generated snapshots found for validation.")
            continue
        test_json_data['snapshot'] = snapshot_key
        testsets = get_field_value_with_default(test_json_data, 'testSet', [])
        for testset in testsets:
            testcases = get_field_value_with_default(testset, 'cases', [])
            testset['cases'] = _get_new_testcases(testcases, mastersnapshots, snapshot_key, container, filesystem=True)
        # print(json.dumps(test_json_data, indent=2))
        # singletest = get_from_currentdata(SINGLETEST)
        # if singletest:
        #     for testset in testsets:
        #         newtestcases = []
        #         for testcase in testset['cases']:
        #             if ('testId' in testcase and  testcase['testId'] == singletest) or \
        #                     ('masterTestId' in testcase and testcase['masterTestId'] == singletest):
        #                 newtestcases.append(testcase)
        #         testset['cases'] = newtestcases
        resultset = run_json_validation_tests(test_json_data, container, True, snapshot_status, dirpath=dirpath)
        if test_json_data.get('testSet') and not resultset:
            logger.error('\tERROR: Testset does not contains any testcases or all testcases are skipped due to invalid rules.')
        elif resultset:
            snapshot = test_json_data['snapshot'] if 'snapshot' in test_json_data else ''
            # if singletest:
            #     print(json.dumps(resultset, indent=2))
            # else:
            #     dump_output_results(resultset, container, test_file, snapshot, True)
            dump_output_results(resultset, container, test_file, snapshot, True)
            for result in resultset:
                if 'result' in result:
                    if not re.match(r'passed', result['result'], re.I):
                        finalresult = False
                        break
        else:
            logger.error('\tERROR: No mastertest Documents found!')
            finalresult = False
    logger.critical("VALIDATION COMPLETE:")
    return finalresult

def run_filecontent_validation(container, snapshot_status=None):
    logger.info("FILE CONTENT VALIDATION:")
    logger.info("\tCollection: %s,  Type: FILESYSTEM", container)
    reporting_path = config_value('REPORTING', 'reportOutputFolder')
    json_dir = '%s/%s/%s' % (framework_dir(), reporting_path, container)
    logger.info('\tLOCATION: %s', json_dir)
    test_files = get_json_files(json_dir, MASTERTEST)
    if not test_files:
        logger.error("ERROR: No `test` or `mastertest` file found. collection should contain either `test` or `mastertest` file")
        return False

    finalresult = True
    for test_file in test_files:
        logger.info('\tCOLLECTION: %s', test_file)
        # logger.info("*" * 50)
        # logger.info("validator tests: %s", test_file)
        dirpath = None
        test_json_data = json_from_file(test_file)
        if not test_json_data:
            logger.info("Test file %s looks to be empty, next!...", test_file)
            continue
        snapshot_key = '%s_gen' % test_json_data['masterSnapshot']
        test_json_data['snapshot'] = snapshot_key
        mastersnapshots = defaultdict(list)
        snapshot_data = snapshot_status[snapshot_key] if snapshot_key in snapshot_status else {}
        for snapshot_id, mastersnapshot_id in snapshot_data.items():
            if isinstance(mastersnapshot_id, list):
                for master_snapshot_id in mastersnapshot_id:
                    mastersnapshots[master_snapshot_id].append(snapshot_id)
            elif isinstance(mastersnapshot_id, str):
                mastersnapshots[mastersnapshot_id].append(snapshot_id)
        testsets = get_field_value_with_default(test_json_data, 'testSet', [])
        for testset in testsets:
            testcases = get_field_value_with_default(testset, 'cases', [])
            # if testIds:
            #     newTestcases = []
            #     for testcase in testcases:
            #         if testcase['masterTestId'] in testIds:
            #             newTestcases.append(testcase)
            #     testcases = newTestcases
            # testset['cases'] = _get_new_testcases(testcases, mastersnapshots)
            testset['cases'] = _get_new_testcases(testcases, mastersnapshots, snapshot_key, container, filesystem=True)


        resultset = run_json_validation_tests(test_json_data, container, filesystem=True, snapshot_status=snapshot_status)
        if test_json_data.get('testSet') and not resultset:
            logger.error('\tERROR: Testset does not contains any testcases or all testcases are skipped due to invalid rules.')
        elif resultset:
            snapshot = test_json_data['snapshot'] if 'snapshot' in test_json_data else ''
            dump_output_results(resultset, container, test_file, snapshot, True)
            for result in resultset:
                if 'result' in result:
                    if not re.match(r'passed', result['result'], re.I):
                        finalresult = False
                        break
        else:
            logger.error('\tERROR: No mastertest Documents found!')
            finalresult = False
    logger.critical("VALIDATION COMPLETE:")
    return finalresult

def _get_snapshot_type_map(container):
    dbname = config_value(DATABASE, DBNAME)
    collection = config_value(DATABASE, collectiontypes[SNAPSHOT])
    qry = {'container': container}
    docs = get_documents(collection, dbname=dbname, query=qry)
    mappings = {}
    if docs and len(docs):
        for doc in docs:
            given_data = doc['json']
            if given_data:
                snapshots = given_data.get("snapshots", [])
                for snapshot in snapshots:
                    given_type = snapshot.get("type","")
                    if given_type == "aws":
                        nodes = snapshot.get("nodes",[])
                        for node in nodes:
                            mappings[node['snapshotId']] = node['type']
    return mappings


def run_container_validation_tests_database(container, snapshot_status=None):
    """ Get the test files from the database"""
    dirpath = None
    dbname = config_value(DATABASE, DBNAME)
    test_files_found = True
    mastertest_files_found = True
    # For test files
    collection = config_value(DATABASE, collectiontypes[TEST])
    qry = {'container': container}
    sort = [sort_field('timestamp', False)]
    docs = get_documents(collection, dbname=dbname, sort=sort, query=qry)
    finalresult = True
    if docs and len(docs):
        logger.info('Number of test Documents: %s', len(docs))
        for doc in docs:
            if doc['json']:
                try:
                    snapshot = doc['json']['snapshot'] if 'snapshot' in doc['json'] else ''
                    test_file = doc['name'] if 'name' in doc else ''
                    update_output_testname(test_file, snapshot, filesystem=False)
                    if "connector" in doc['json'] and "remoteFile" in doc['json'] and doc['json']["connector"] and doc['json']["remoteFile"]:
                        dirpath, pull_response = pull_json_data(doc['json'])
                        if not pull_response:
                            return {}
                    resultset = run_json_validation_tests(doc['json'], container, False, dirpath=dirpath)
                    if resultset:
                        # dump_output_results(resultset, container, test_file, snapshot, False)
                        for result in resultset:
                            if 'result' in result:
                                if not re.match(r'passed', result['result'], re.I):
                                    finalresult = False
                                    break
                except Exception as e:
                    # dump_output_results([], container, "-", snapshot, False)
                    raise e
    else:
        logger.info('No test Documents found!')
        test_files_found = False
        finalresult = False
    # For mastertest files
    collection = config_value(DATABASE, collectiontypes[MASTERTEST])
    docs = get_documents(collection, dbname=dbname, sort=sort, query=qry)
    # snapshots_details_map = _get_snapshot_type_map(container)
    if docs and len(docs):
        logger.info('Number of mastertest Documents: %s', len(docs))
        for doc in docs:
            test_json_data = doc['json']
            if test_json_data:
                snapshot = doc['json']['masterSnapshot'] if 'masterSnapshot' in doc['json'] else ''
                test_file = doc['name'] if 'name' in doc else '-'
                update_output_testname(test_file, snapshot, filesystem=False)
                try:
                    if "connector" in test_json_data and "remoteFile" in test_json_data and test_json_data["connector"] and test_json_data["remoteFile"]:
                        dirpath, pull_response = pull_json_data(test_json_data)
                        if not pull_response:
                            return {}
                    snapshot_key = '%s_gen' % test_json_data['masterSnapshot']
                    mastersnapshots = defaultdict(list)
                    snapshot_data = snapshot_status[snapshot_key] if snapshot_key in snapshot_status else {}
                    for snapshot_id, mastersnapshot_id in snapshot_data.items():
                        if isinstance(mastersnapshot_id, list):
                            for msnp_id in mastersnapshot_id:
                                mastersnapshots[msnp_id].append(snapshot_id)    
                        else:
                            mastersnapshots[mastersnapshot_id].append(snapshot_id)
                    test_json_data['snapshot'] = snapshot_key
                    testsets = get_field_value_with_default(test_json_data, 'testSet', [])
                    for testset in testsets:
                        testcases = get_field_value_with_default(testset, 'cases', [])
                        testcases += COMPLIANCES
                        testset['cases'] = _get_new_testcases(testcases, mastersnapshots, snapshot_key, container, dbname, False)
                    # print(json.dumps(test_json_data, indent=2))
                    resultset = run_json_validation_tests(test_json_data, container, False, snapshot_status, dirpath=dirpath)
                    if resultset:
                        # dump_output_results(resultset, container, test_file, snapshot, False)
                        for result in resultset:
                            if 'result' in result:
                                if not re.match(r'passed', result['result'], re.I):
                                    finalresult = False
                                    break
                except Exception as e:
                    # dump_output_results([], container, test_file, snapshot, False)
                    raise e
    else:
        logger.info('No mastertest Documents found!')
        mastertest_files_found = False
        finalresult = False
    if not test_files_found and not mastertest_files_found:
        raise Exception("No complaince tests for this container: %s, add and run!", container)
    return finalresult


def _get_new_testcases(testcases, mastersnapshots, snapshot_key, container, dbname=None, filesystem=True):
    onlysnapshots = get_from_currentdata("ONLYSNAPSHOTS")
    onlysnapshotsIds = get_from_currentdata("ONLYSNAPSHOTIDS")
    includeTestConfig = get_from_currentdata("INCLUDETESTS")
    includeTests = get_from_currentdata("TESTIDS")

    excludedTestIds = []
    exclusions = get_from_currentdata(EXCLUSION).get('exclusions', [])
    for exclusion in exclusions:
        if 'exclusionType' in exclusion and exclusion['exclusionType'] and exclusion['exclusionType'] == 'test':
            if 'masterTestID' in exclusion and exclusion['masterTestID'] and exclusion['masterTestID'] not in excludedTestIds:
                if includeTestConfig:
                    if exclusion['masterTestID'] not in includeTests:
                        excludedTestIds.append(exclusion['masterTestID'])
                else:
                    excludedTestIds.append(exclusion['masterTestID'])
    
    snapshot_json_data = get_snapshot_file(snapshot_key, container, dbname, filesystem)
    snapshots = snapshot_json_data.get("snapshots")
    snapshot_resource_map = {}
    for snapshot in snapshots:
        for node in snapshot["nodes"]:
            snapshot_resource_map[node.get("snapshotId")] = node.get("resourceTypes", [])
    
    newcases = []
    for testcase in testcases:
        if 'masterTestId' in testcase and testcase['masterTestId'] and testcase['masterTestId'] in excludedTestIds:
            logger.warning("Excluded from testId exclusions: %s", testcase['masterTestId'])
            continue
        if includeTestConfig:
            if 'masterTestId' in testcase and testcase['masterTestId'] and testcase['masterTestId'] not in includeTests:
                continue

        test_parser_type = testcase.get('type', None)
        if test_parser_type == 'rego' or test_parser_type == 'python':
            new_cases = _get_rego_testcase(testcase, mastersnapshots, snapshot_resource_map)
            newcases.extend(new_cases)
        else:
            rule_str = get_field_value_with_default(testcase, 'rule', '')
            ms_ids = re.findall(r'\{(.*)\}', rule_str)
            # detail_method = get_field_value(testcase, 'detailMethod')
            for ms_id in ms_ids:
                for s_id in mastersnapshots[ms_id]:
                    # new_rule_str = re.sub('{%s}' % ms_id, '{%s}' % s_id, rule_str)
                    # if not detail_method or detail_method == snapshots_details_map[s_id]:
                    toAdd = True
                    if onlysnapshots:
                        if s_id not in onlysnapshotsIds:
                            toAdd = False
                    if toAdd:
                        new_rule_str = rule_str.replace('{%s}' % ms_id, '{%s}' % s_id)
                        new_testcase = {
                            'title': testcase.get('title') if testcase.get('title') else "",
                            'description': testcase.get('description') if testcase.get('description') else "",
                            'rule': new_rule_str,
                            'testId': testcase['masterTestId'],
                            'status' : get_field_value_with_default(testcase, 'status', "enable")
                        }
                        newcases.append(new_testcase)
    return newcases

def _get_rego_testcase(testcase, mastersnapshots, snapshot_resource_map):
    onlysnapshots = get_from_currentdata("ONLYSNAPSHOTS")
    onlysnapshotsIds = get_from_currentdata("ONLYSNAPSHOTIDS")
    newcases = []
    ms_ids = testcase.get('masterSnapshotId')
    # service = ms_ids[0].split('_')[1]
    for ms_id in ms_ids:
        if ms_id == "ALL":
            for ms_id, s_ids in mastersnapshots.items():
                for s_id in s_ids:
                    snapshot_resource_types = snapshot_resource_map.get(s_id, [])
                    testcase_resource_types = testcase.get('resourceTypes', [])
                    
                    if testcase_resource_types and all(resource_type not in snapshot_resource_types for resource_type in testcase_resource_types):
                        continue

                    toAdd = True
                    if onlysnapshots:
                        if s_id not in onlysnapshotsIds:
                            toAdd = False
                    if toAdd:
                        new_testcase = copy.copy(testcase)
                        new_testcase['snapshotId'] = [s_id]
                        newcases.append(new_testcase)
        else:
            for s_id in mastersnapshots[ms_id]:
                snapshot_resource_types = snapshot_resource_map.get(s_id, [])
                testcase_resource_types = testcase.get('resourceTypes', [])
                
                if testcase_resource_types and all(resource_type not in snapshot_resource_types for resource_type in testcase_resource_types):
                    continue

                toAdd = True
                if onlysnapshots:
                    if s_id not in onlysnapshotsIds:
                        toAdd = False
                if toAdd:
                    new_testcase = copy.copy(testcase)
                    # if service not in ms_id:
                    #     if s_id.split('_')[1] not in newcases[0]['snapshotId'][-1]:
                    #         [newcase['snapshotId'].append(s_id) for newcase in newcases]
                    #     else:
                    #         new_cases = copy.deepcopy(newcases)
                    #         [new_case['snapshotId'].pop(-1) and new_case['snapshotId'].append(s_id) for new_case in new_cases]
                    #         newcases.extend(new_cases)
                    #     continue
                    new_testcase['snapshotId'] = [s_id]
                    newcases.append(new_testcase)
    return newcases

def container_snapshots_filesystem(container):
    """Get snapshot list used in test files from the filesystem."""
    snapshots = []
    logger.info("Starting to get list of snapshots")
    reporting_path = config_value('REPORTING', 'reportOutputFolder')
    json_dir = '%s/%s/%s' % (framework_dir(), reporting_path, container)
    logger.info(json_dir)
    test_files = get_json_files(json_dir, JSONTEST)
    logger.info('\n'.join(test_files))
    for test_file in test_files:
        test_json_data = json_from_file(test_file)
        if test_json_data:
            snapshot = test_json_data['snapshot'] if 'snapshot' in test_json_data else ''
            if snapshot:
                snapshots.append(snapshot)
    return snapshots

def main(container):
    run_container_validation_tests_filesystem(container)


if __name__ == '__main__':
    import sys
    from processor.logging.log_handler import getlogger, init_logger, NONE
    init_logger(NONE)
    if len(sys.argv) > 1:
        main(sys.argv[1])



"""
Run time data storage and retrieval.
"""
import time
import datetime
import json
import socket
import os.path
import shutil
from processor.helper.config.config_utils import config_value, framework_currentdata, get_cache_data, \
    set_cache_data, TESTS, DBTESTS, DBVALUES, SNAPSHOT
from processor.helper.json.json_utils import json_from_file, save_json_to_file
from processor.logging.log_handler import getlogger, FWLOGFILENAME
from processor.helper.file.file_utils import remove_file, exists_dir, mkdir_path

exclude_list = ['token', 'clientSecret', 'vaulttoken', 'exclusion', 'apitoken', 'gittoken', 'outputpath']

def get_dbtests():
    currdata = get_currentdata()
    if DBTESTS in currdata:
        dbtests = currdata[DBTESTS]
    else:
        nodb = config_value(TESTS, DBTESTS)
        if nodb and nodb.upper() in DBVALUES:
            dbtests = DBVALUES.index(nodb.upper())
        else:
            dbtests = DBVALUES.index(SNAPSHOT)
        put_in_currentdata(DBTESTS, dbtests)
    return dbtests



def add_to_exclude_list(key):
    if key not in exclude_list:
        exclude_list.append(key)


def init_currentdata():
    """ Initialises data structure to store runtime data. """
    started = int(time.time() * 1000)
    runctx = framework_currentdata()
    run_dir = os.path.dirname(runctx)
    if not exists_dir(run_dir):
        mkdir_path(run_dir)
    run_data = {
        'start': started,
        'end': started,
        'remote': False,
        'errors': [],
        'host': socket.gethostname(),
        'timestamp': datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
    }
    save_currentdata(run_data)


def put_in_currentdata(key, value):
    """Adds a value in the current run data"""
    if key:
        curr_data = get_currentdata()
        if key in curr_data:
            val = curr_data[key]
            if isinstance(val, list):
                val.append(value)
            else:
                curr_data[key] = value
        else:
            curr_data[key] = value
        save_currentdata(curr_data)

def delete_from_currentdata(key):
    """Remove a key from the current run data"""
    if key:
        currdata = get_currentdata()
        if key in currdata:
            del currdata[key]
        save_currentdata(currdata)


def get_from_currentdata(key):
    """ Get the data for this key from the rundata"""
    data = None
    currdata = get_currentdata()
    if key and key in currdata:
        data = currdata[key]
    return data


def get_currentdata():
    """Get the current run data, if present else empty json object"""
    runctx = framework_currentdata()
    curr_data = json_from_file(runctx)
    if not curr_data:
        curr_data = {}
    return curr_data

def put_in_cachedata(key, value):
    """Adds a value in the cache data"""
    if key:
        curr_data = get_cache_data()
        if key in curr_data:
            val = curr_data[key]
            if isinstance(val, list):
                val.append(value)
            else:
                curr_data[key] = value
        else:
            curr_data[key] = value
        set_cache_data(curr_data)

def get_from_cachedata(key):
    """ Get the data for this key from the cachedata"""
    data = None
    currdata = get_cache_data()
    if key and key in currdata:
        data = currdata[key]
    return data


def save_currentdata(curr_data):
    """Save the key value rundata for further access, if None store it empty."""
    if not curr_data:
        curr_data = {}
    runctx = framework_currentdata()
    save_json_to_file(curr_data, runctx)


def delete_currentdata():
    """Delete the rundata file when exiting of the script."""
    logger = getlogger()
    # singletest = get_from_currentdata(SINGLETEST)
    # if singletest:
    #     container = get_from_currentdata('container')
    #     cdir = get_container_dir(container)
    #     shutil.rmtree('%s/snapshots' % cdir)

    cleaning_repos = get_from_currentdata("CLEANING_REPOS")
    if cleaning_repos:
        for repo in cleaning_repos:
            if repo and os.path.exists(repo):
                shutil.rmtree(repo)
        delete_from_currentdata("CLEANING_REPOS")

    logger.info("END: Completed the run and cleaning up.")
    runctx = get_currentdata()
    runctx['end'] = int(time.time() * 1000)
    runctx['log'] = FWLOGFILENAME
    if 'start' in runctx:
        runctx['duration'] = '%d seconds' % int((runctx['end'] - runctx['start'])/1000)
        runctx['start'] = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(runctx['start']/1000))
    runctx['end'] = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(runctx['end']/1000))
    newdata = {}
    for k,v in runctx.items():
        if k not in exclude_list:
            newdata[k] = v
    logger.critical("\033[92m Run Stats: %s\033[00m" % json.dumps(newdata, indent=2))
    if runctx['remote']:
        from processor.helper.utils.compliance_utils import upload_compliance_results
        logger.info("Uploading data....")
        upload_compliance_results(runctx['container'], runctx['outputpath'], runctx['env'], runctx['company'], runctx['apitoken'])
    run_file = framework_currentdata()
    remove_file(run_file)





""" Utility functions for yaml."""

import yaml
from yaml.loader import FullLoader
from collections import OrderedDict
from processor.helper.file.file_utils import exists_file
from processor.logging.log_handler import getlogger

logger = getlogger()
MultipleConvertionKey = "_multiple_yaml"
HelmChartConvertionKey = "_prancer_helm_template"

def save_yaml_to_file(indata, outfile, indent=None):
    """Save dict data to the file in yaml format"""
    if indata is not None:
        try:
            with open(outfile, 'w') as yamlfile:
                yaml.dump(indata, yamlfile, indent=indent)
        except:
            pass


def yaml_from_string(yaml_str):
    """Get dict from the string in yaml format."""
    try:
        yamldata = yaml.load(yaml_str, Loader=FullLoader)
        return yamldata
    except:
        print('Failed to load yaml data: %s' % yaml_str)
    return None


def yaml_from_file(yamlfile, loader=None):
    """ Get yaml data from the file in a dict."""
    yamldata = None
    try:
        if exists_file(yamlfile):
            with open(yamlfile) as infile:
                if loader:
                    yamldata = yaml.load(infile, Loader=loader)
                else:
                    yamldata = yaml.load(infile, Loader=FullLoader)
    except Exception as ex:
        print('Failed to load yaml from file: %s, exception: %s' % (yamlfile, ex))
    return yamldata


def valid_yaml(yaml_input):
    """ Checks validity of the yaml """
    try:
        data = yaml.load(yaml_input)
        return isinstance(data, dict)
    except:
        print('Not a valid yaml: %s' % yaml_input)
    return False

def multiple_yaml_from_file(yamlfile, loader=None):
    """ Get multiple yaml data from the file in a dict."""
    yamldata = None
    try:
        if exists_file(yamlfile):
            with open(yamlfile) as infile:
                if loader:
                    yamldata = list(yaml.load_all(infile, Loader=loader))
                else:
                    yamldata = list(yaml.load_all(infile))
    except Exception as ex:
        return None
    return yamldata

def is_multiple_yaml_file(file_path):
    try:
      if len (multiple_yaml_from_file(file_path,loader=FullLoader)) > 1:
          return True
      else: 
          return False
    except Exception as ex:
        return False

def is_multiple_yaml_convertion(file_path):
    return MultipleConvertionKey in file_path

def is_helm_chart_convertion(file_path):
    return HelmChartConvertionKey in file_path
    


"""Reporting related utility functions."""
import hashlib
import time
from bson.objectid import ObjectId
from processor.helper.config.config_utils import config_value
from collections import OrderedDict
from processor.helper.json.json_utils import save_json_to_file, collectiontypes, OUTPUT
from processor.database.database import DATABASE, DBNAME, find_and_update_document, get_documents, insert_one_document, update_one_document
from processor.logging.log_handler import get_dblogger

doc_id = None
dbname = None
collection = None

def json_record(container, filetype, filename, json_data=None):
    db_record = {
        "timestamp": int(time.time() * 1000),
        "container": container,
        "checksum": hashlib.md5("{}".encode('utf-8')).hexdigest(),
        "type": filetype,
        "name": filename,
        "collection": config_value(DATABASE, collectiontypes[filetype]),
        "json": json_data if json_data else {}
    }
    if '$schema' in db_record['json']:
        del db_record['json']['$schema']
    return db_record

def create_output_entry(container, test_file="", filesystem=False):
    global doc_id, dbname, collection
    od = OrderedDict()
    od["$schema"] = ""
    od["contentVersion"] = "1.0.0.0"
    od["fileType"] = OUTPUT
    od["timestamp"] = int(time.time() * 1000)
    od["container"] = container
    od["status"] = "Running"
    dblog = get_dblogger()
    od["log"] = dblog if dblog else ""
    if not filesystem:
        od["test"] = test_file
        od["results"] = []
        del od["$schema"]
        doc = json_record(container, OUTPUT, test_file, od)
        dbname = config_value(DATABASE, DBNAME)
        collection = config_value(DATABASE, collectiontypes[OUTPUT])
        doc_id = insert_one_document(doc, collection, dbname)
    return doc_id

def update_output_testname(test_file="", snapshot="", filesystem=False):
    if doc_id and not filesystem:
        find_and_update_document(
            collection=collection,
            dbname=dbname,
            query={"_id" : ObjectId(doc_id)},
            update_value={
                "$set" :  {
                    "json.test": test_file,
                    "json.snapshot" : snapshot
                }
            }
        )

def dump_output_results(results, container, test_file, snapshot, filesystem=True, status=None):
    """ Dump the report in the json format for test execution results."""
    dbname = config_value(DATABASE, DBNAME)
    collection = config_value(DATABASE, collectiontypes[OUTPUT])
    if not doc_id:
        od = OrderedDict()
        od["$schema"] = ""
        od["contentVersion"] = "1.0.0.0"
        od["fileType"] = OUTPUT
        od["timestamp"] = int(time.time() * 1000)
        od["snapshot"] = snapshot
        od["container"] = container
        dblog = get_dblogger()
        od["log"] = dblog if dblog else ""
        if filesystem:
            test_file_parts = test_file.rsplit('/', 1)
            od["test"] = test_file_parts[-1]
            output_file = '%s/output-%s' % (test_file_parts[0], test_file_parts[-1])
            od["results"] = results
            save_json_to_file(od, output_file)
        else:
            od["test"] = test_file
            od["results"] = results
            del od["$schema"]
            doc = json_record(container, OUTPUT, test_file, od)
            insert_one_document(doc, collection, dbname)
    else:
        update_value = {}
        if results:
            update_value["$push"] = { "json.results": { "$each" : results }}
        
        if status:
            update_value["$set"] = { "json.status": status }
            
        find_and_update_document(
            collection=collection,
            dbname=dbname,
            query={"_id" : ObjectId(doc_id)},
            update_value=update_value
        )



"""A setup module for prancer-basic."""


# setuptools for distribution
from setuptools import find_packages, setup
import os
from src import processor


with open('requirements.txt') as f:
    required = f.read().splitlines()

LONG_DESCRIPTION = """
 Prancer Basic allows to users to run cloud validation.
 The supported cloud frameworks are azure, aws and git.
"""

setup(
    name='prancer-basic',
    # also update the version in processor.__init__.py file
    version='2.0.13',
    description='Prancer Basic, http://prancer.io/',
    long_description=LONG_DESCRIPTION,
    license = "BSD",
    # The project's main homepage.
    url='https://github.com/prancer-io/cloud-validation-framework',
    # Author(s) details
    author='Farshid M/Ajey Khanapuri',
    author_email='ajey.khanapuri@liquware.com',
    classifiers=[
        "Development Status :: 3 - Alpha",
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
        "License :: OSI Approved :: BSD License",
    ],
    packages=find_packages(where="src",
                           exclude=['log', 'rundata', 'utilities', 'tests']),
    include_package_data=True,
    package_dir={'': 'src'},
    setup_requires=['ply==3.10'],
    install_requires=required,
    python_requires='>=3.0',
    entry_points={
        'console_scripts': [
            'validator = processor.helper.utils.cli_validator:validator_main',
            'prancer = processor.helper.utils.cli_validator:validator_main',
            'populate_json = processor.helper.utils.cli_populate_json:populate_json_main',
            'terraform_to_json = processor.helper.utils.cli_terraform_to_json:terraform_to_json_main',
            'register_key_in_azure_vault = processor.helper.utils.cli_generate_azure_vault_key:generate_azure_vault_key'
        ],
    }
)




import json
from hcl.parser import HclParser, pickle_file
from hcl.api import u, isHcl
from hcl2.lark_parser import Lark_StandAlone
from processor.helper.hcl import yacc
from processor.helper.hcl.transformer import HClDictTransformer

class TerraformHCLParer(HclParser):
    def __init__(self):
        self.yacc = yacc.yacc(module=self, debug=False, optimize=1, picklefile=pickle_file)
    
    def parse(self, s):
        return self.yacc.parse(s, lexer=yacc.TerraformLexer())

def loads(fp):
    '''
    Deserializes a string and converts it to a dictionary. The contents
    of the string must either be JSON or HCL.
    
    :returns: Dictionary 
    '''
    s = fp.read()
    hcl2 = Lark_StandAlone(transformer=HClDictTransformer())
    return hcl2.parse(s + "\n")
    # s = u(s)
    # if isHcl(s):
    #     return TerraformHCLParer().parse(s)
    # else:
    #     return json.loads(s)



class BaseCrawler:
    def __init__(self, resources, **kwargs):
        self.resources = resources
        self.resource_types = {}

    def check_for_special_crawl(self, resource_type):
        """ 
        check the resource type need special crawling or not. if it need special crawling then 
        it will process the special crawler and returns the updated resources
        """
        return self.resources


"""
   Common utility file to convert terraform to json files.
"""
import argparse
import sys
import atexit
from processor.logging.log_handler import getlogger
from processor.helper.config.rundata_utils import init_currentdata, delete_currentdata
from processor.helper.file.file_utils import exists_file
from processor.helper.json.json_utils import save_json_to_file
from processor.connector.snapshot_custom import convert_to_json




def convert_terraform_to_json(terraform, output=None):
    if exists_file(terraform):
        if not output:
            parts = terraform.rsplit('.', -1)
            output = '%s.json' % parts[0]
        _, json_data = convert_to_json(terraform, 'terraform')
        if json_data:
            save_json_to_file(json_data, output)


def terraform_to_json_main(arg_vals=None):
    """Main driver utility for converting terraform to json files."""
    logger = getlogger()
    logger.info("Comand: '%s %s'", sys.executable.rsplit('/', 1)[-1], ' '.join(sys.argv))
    cmd_parser = argparse.ArgumentParser("Convert terraform to json files")
    cmd_parser.add_argument('terraform', action='store',
                            help='Full path of the terraform file.')
    cmd_parser.add_argument('--output', action='store', default=None,
                            help='Path to store the file.')
    args = cmd_parser.parse_args(arg_vals)
    # Delete the rundata at the end of the script.
    atexit.register(delete_currentdata)
    logger.info(args)
    init_currentdata()
    convert_terraform_to_json(args.terraform, args.output)
    return 0



"""
Define an interface for parsing aws template and its parameter files.
"""
import json
import copy
from cfn_flip import to_json
from processor.helper.json.json_utils import json_from_file
from processor.helper.file.file_utils import exists_file
from processor.logging.log_handler import getlogger
from processor.templates.base.template_parser import TemplateParser

logger = getlogger()
allowed_extensions = ["json","yaml", "template", "txt"]

class AWSTemplateParser(TemplateParser):
    """
    AWS Parser class for process AWS cloudformation template
    """
    def __init__(self, template_file, tosave=False, **kwargs):
        super().__init__(template_file, tosave=False, **kwargs)
        self.mappings = {}
        self.intrinsic_functions = {
            "FindInMap" : self.handle_find_in_map,
            "Join" : self.handle_join,
            "If": self.handle_if,
            "Equals": self.handle_equals,
            "And": self.handle_and,
            "Or": self.handle_or,
            "Not": self.handle_not,
            "GetAtt": self.handle_get_att,
            "Select": self.handle_select,
            "Split": self.handle_split,
            "Sub": self.handle_sub,
        }
    
    def yaml_to_json(self, yaml_file):
        """
        takes the yaml file path and converts the returns the converted JSON object
        """
        template_json = None
        with open(yaml_file, encoding="utf-8") as yml_file:
            try:
                template_json = json.loads(to_json(yml_file.read()))
            except:
                file_name = yaml_file.split("/")[-1]
                logger.error("Failed to load yaml file, please check yaml file contains correct content: %s", file_name)
        return template_json
    
    def generate_template_json(self):
        """
        generate the template json from template and parameter files
        """
        gen_template_json = None
        template_json = None
        if self.get_template().endswith(".yaml") and exists_file(self.get_template()):
            template_json = self.yaml_to_json(self.get_template())
            self.contentType = 'yaml'
        elif self.get_template().endswith(".json") :
            template_json = json_from_file(self.get_template(), object_pairs_hook=None)
        elif self.get_template().endswith(".template") or self.get_template().endswith(".txt"):
            template_json = json_from_file(self.get_template(), object_pairs_hook=None)
            if template_json:
                self.contentType = "json"
            else:
                try:
                    template_json = self.yaml_to_json(self.get_template())
                    self.contentType = 'yaml'
                except:
                    pass

        self.template_json = template_json
        if not template_json:
            logger.error("Invalid path! No file found at : %s", self.get_template())
            return gen_template_json

        if "AWSTemplateFormatVersion" not in template_json:
            logger.error("Invalid file content : file does not contains 'AWSTemplateFormatVersion' field.")
            return gen_template_json

        if template_json:
            gen_template_json = copy.deepcopy(template_json)
            if 'Parameters' in template_json:
                self.gparams = template_json['Parameters']
                if self.parameter_file:
                    parameters = json_from_file(self.parameter_file, object_pairs_hook=None)
                    # parameters = self.json_from_file(self.parameter_file)
                    if parameters:
                        for param in parameters:
                            if "ParameterKey" in param and "ParameterValue" in param:
                                self.gparams[param["ParameterKey"]] = { "Default" : param["ParameterValue"] } 
            if 'Mappings' in template_json:
                self.mappings = template_json['Mappings']
            if 'Resources' in template_json:
                new_resources = []
                for key, resource in template_json['Resources'].items():
                    new_resource = self.process_resource(resource)
                    if "Type" in new_resource:
                        self.resource_types.append(new_resource.get("Type").lower())
                    new_resource["Name"] = key
                    new_resources.append(new_resource)
                    gen_template_json['Resources'] = new_resources
        return gen_template_json

    def process_function(self, resource):
        """
        performs the Intrinsic function on given resource
        """
        value = resource
        if isinstance(resource ,dict):
            if "Ref" in resource:
                value = self.handle_reference(value)
            else: 
                keys = value.keys()
                function = None
                for k in keys:
                    if k.startswith("Fn::"):
                        function = k
                if function:
                    function = function.split("Fn::")[1]
                    if function in self.intrinsic_functions:
                        value = self.intrinsic_functions[function](value)
                else:
                    value = self.process_resource(value)
        return value
    
    def process_resource(self, resource):
        """ 
        process the resource json and return the resource with updated values
        """
        new_resource = resource
        if isinstance(resource ,dict):
            new_resource = {}
            for key, value in resource.items():
                if isinstance(value, dict):
                    new_resource[key] = self.process_function(value)
                else:
                    result = self.process_resource(value)
                    new_resource[key] = result
        elif isinstance(resource ,list):
            new_resource = []
            for value in resource:
                value = self.process_function(value)
                if isinstance(value, str):
                    new_resource.append(value)
                else:
                    new_resource.append(self.process_resource(value))
        return new_resource
    
    def handle_reference(self, value):
        """
        Returns the default value for specified parameter reference
        """
        if value["Ref"] in self.gparams and "Default" in self.gparams[value["Ref"]]:
            value = self.gparams[value["Ref"]]["Default"]
        return value
    
    def handle_find_in_map(self, value):
        """
        Finding the appropriate value from the Mapping 
        """
        find_values = value["Fn::FindInMap"]

        for item in find_values:
            item = self.process_resource(item)

        if len(find_values) == 3 and all(isinstance(v, str) for v in find_values):
            map_name = find_values[0]
            top_level_key = find_values[1]
            second_level_key = find_values[2]

            if map_name in self.mappings and top_level_key in self.mappings[map_name] and \
                second_level_key in self.mappings[map_name][top_level_key]:
                value = self.mappings[map_name][top_level_key][second_level_key]        
        return value
    
    def handle_join(self, value):
        """
        Performes the `join` operation on set of values with specified delimiter.
        """
        join_values = value["Fn::Join"]
        for i in range(0, len(join_values)):
            join_values[i] = self.process_resource(join_values[i])
        
        if len(join_values) == 2 and all(isinstance(v, str) for v in join_values[1]):
            value = join_values[0].join(join_values[1])
        if not isinstance(value, str):
            value = self.process_resource(value)
        
        return value

    def handle_if(self, value):
        """
        perform the 'If' operation on values
        """
        condition, true_value, false_value = value["Fn::If"]
        new_condition = self.handle_condition(condition)
        true_value = self.process_resource(true_value)
        false_value = self.process_resource(false_value)

        
        if new_condition != condition:
            if new_condition == True:
                return true_value
            elif new_condition == False:
                return false_value
        else:
            return value

    def handle_equals(self, value):
        """
        Handle equals between 2 values
        """
        first_value, second_value = value.get("Fn::Equals")

        updated_first_value = self.process_handler_value(first_value)
        updated_second_value = self.process_handler_value(second_value)
        

        if updated_first_value == updated_second_value:
            return True
        else:
            return False
    
    def handle_and(self, value):
        """
        Handle AND condition within list
        """
        condition_list = value["Fn::And"]
        updated_condition_list = []
        for condition in condition_list:
            updated_condition = self.process_handler_value(condition)
            updated_condition_list.append(updated_condition)
            if not isinstance(updated_condition, bool):
                return value

        return all(updated_condition_list)
    
    def handle_or(self, value):
        """
        Handle Or condition within list
        """
        condition_list = value["Fn::Or"]
        updated_condition_list = []
        for condition in condition_list:
            updated_condition = self.process_handler_value(condition)
            updated_condition_list.append(updated_condition)
            if not isinstance(updated_condition, bool):
                return value

        return any(updated_condition_list)
    
        
    def handle_not(self, value):
        """
        Handle Not condition
        """
        not_value = value["Fn::Not"][0]
        updated_not_value = self.process_handler_value(not_value)

        if isinstance(updated_not_value, bool):
            return not updated_not_value
        else:
            return value
    
    def handle_get_att(self, value):
        """
        get attribute from local resource and update property
        """
        resource_name, attr_name = value["Fn::GetAtt"]
        attr_path_list = attr_name.split(".")
        for resource in self.template_json.get("Resources", []):
            try:
                if resource.get("Name") == resource_name:
                    resource_properties = resource.get("Properties",{})
                    resource_properties = self.process_handler_value(resource_properties)
                    for attr in attr_path_list:
                        resource_properties = resource_properties.get(attr,{})
                        if resource_properties == None:
                            return value
                    return resource_properties
            except:
                return value
        return value
    
    def handle_select(self, value):
        """
        select indexed value from list and update property
        """
        index, list_value = value["Fn::Select"]
        index = int(index)
        list_value = self.process_handler_value(list_value)
        if isinstance(list_value, list) and len(list_value)-1 >= index:
            return list_value[index]
        else:
            return value
    
    def handle_split(self, value):
        """
        split string and return list of values
        """
        delimiter, source_string = value["Fn::Split"]
        source_string = self.process_handler_value(source_string)
        if isinstance(source_string, str):
            return source_string.split(delimiter)
        else:
            return value
    
    def handle_sub(self, value):
        """
        substitute dict value in string
        """
        if isinstance(value["Fn::Sub"], list):
            string, sub_value_dict = value["Fn::Sub"]
            for key, val in sub_value_dict.items():
                val = self.process_handler_value(val)
                if string.find("${%s}"%(key)) != -1:
                    string = string.replace("${%s}"%(key),val)
                else:
                    return value
            return string
        else:
            return value
    
    def handle_condition(self, value):
        """
        handle "condition" by getting reference from 
        """
        resource_condition = self.template_json.get("Conditions")
        condition = resource_condition.get(value)
        if condition:
            updated_condition = self.process_function(condition)
            if condition != updated_condition:
                return updated_condition
            else:
                return value
        else:
            return value

    def process_handler_value(self, value):
        def all_keys(dict_obj):
            ''' This function generates all keys of
                a nested dictionary. 
            '''
            # Iterate over all keys of the dictionary
            for key , value in dict_obj.items():
                yield key
                # If value is of dictionary type then yield all keys
                # in that nested dictionary
                if isinstance(value, dict):
                    for k in all_keys(value):
                        yield k
        updated_value = value
        if isinstance(value, dict):
            process_function_value = False
            for key in value.keys():
                if key in ["Ref", "Fn::If", "Fn::Equals", "Fn:And", "Fn::Not", "Fn::Or", "Fn::Split", "Fn::Select", "Fn::GetAtt", "Fn::Sub"]:
                    process_function_value = True
            
            if process_function_value:
                updated_value = self.process_function(value)
            else:
                updated_value = self.process_resource(value)
            
        elif isinstance(value, list):
            updated_value = self.process_resource(value)
        
        else:
            updated_value = value
        
        return updated_value

if __name__ == '__main__':
    parameter_file = None
    template_file = '/tmp/templates/SQS_With_CloudWatch_Alarms.template'
    aws_template_parser = AWSTemplateParser(template_file, parameter_file=parameter_file)
    template_json = aws_template_parser.parse()
    print(json.dumps(template_json, indent=2))



"""
   Common file for vault functionality.
"""
from builtins import input
import os
from subprocess import Popen, PIPE
from processor.logging.log_handler import getlogger
from processor.helper.config.rundata_utils import get_from_currentdata,\
    put_in_currentdata, add_to_exclude_list
from processor.helper.config.config_utils import config_value
from processor.helper.httpapi.restapi_azure import get_vault_access_token, get_uami_vault_access_token,\
    get_keyvault_secret, set_keyvault_secret, get_all_secrets, delete_keyvault_secret, set_keyvault_secret_with_response

logger = getlogger()


def get_vault_data(secret_key=None):
    """Read vault data from config"""
    vaulttype = config_value('VAULT', 'type')
    val = None
    if vaulttype:
        if vaulttype == 'azure':
            val = get_azure_vault_data(secret_key)
        elif vaulttype == 'cyberark':
            val = get_cyberark_data(secret_key)
    return val


def set_vault_data(key_name=None, value=None):
    """Update vault data"""
    vaulttype = config_value('VAULT', 'type')
    val = None
    if vaulttype:
        if vaulttype == 'azure':
            val = set_azure_vault_data(key_name, value)
    return val

def set_vault_data_with_response(key_name=None, value=None):
    """Update vault data"""
    vaulttype = config_value('VAULT', 'type')
    status, response = None, None
    if vaulttype:
        if vaulttype == 'azure':
            status, response = set_azure_vault_data_with_response(key_name, value)
    return status, response

def delete_vault_data(secret_key=None):
    """Delete vault data from config"""
    vaulttype = config_value('VAULT', 'type')
    val = None
    if vaulttype:
        if vaulttype == 'azure':
            val = delete_azure_vault_data(secret_key)
    return val


def get_all_vault_secrets():
    """Read all vault secrets"""
    vaulttype = config_value('VAULT', 'type')
    val = None
    if vaulttype:
        if vaulttype == 'azure':
            val = get_all_azure_secrets()
    return val


def get_all_azure_secrets():
    val = None
    vaulttoken = _get_vault_token()
    # logger.debug('Vault Token: %s', vaulttoken)
    if vaulttoken:
        keyvault = config_value('VAULT', 'keyvault')
        logger.info('Keyvault: %s', keyvault)
        data = get_all_secrets(keyvault, vaulttoken)
        if data:
            return data
    return [] 


def get_config_value(section, key, env_var, prompt_str=None):
    """ Return the client secret used for the current run"""
    client_secret = config_value(section, key)
    if not client_secret and env_var:
        client_secret = os.getenv(env_var, None)
    if not client_secret and prompt_str:
        key_str = '%s_%s' % (section, key)
        client_secret = get_from_currentdata(key_str)
        if not client_secret:
            client_secret = input(prompt_str)
            if client_secret:
                put_in_currentdata(key_str, client_secret)
                logger.info('Key:%s, sec:%s', key_str, '*' * len(client_secret))
                add_to_exclude_list(key_str)
    return client_secret


def _get_vault_token():
    """Fetches the bearer token for Azure Vault API calls"""
    if 'UAMI' in os.environ and os.environ['UAMI'] == 'true':
        vaulttoken = get_uami_vault_access_token()
    else:
        client_id = config_value('VAULT', 'client_id')
        client_secret = get_config_value('VAULT', 'client_secret', 'CLIENTKEY',
                                         'Enter the client secret to access keyvault: ')
        # client_secret = config_value('VAULT', 'client_secret')
        tenant_id = config_value('VAULT', 'tenant_id')
        # logger.info('Id: %s, secret: %s, tenant: %s', client_id, client_secret, tenant_id)
        vaulttoken = get_vault_access_token(tenant_id, client_id, client_secret)
    return vaulttoken


def get_azure_vault_data(secret_key=None):
    """Fetches the bearer token for Azure Vault API calls"""
    val = None
    vaulttoken = _get_vault_token()
    # logger.debug('Vault Token: %s', vaulttoken)
    if vaulttoken and secret_key:
        keyvault = config_value('VAULT', 'keyvault')
        # secret_key = config_value('VAULT', 'secret_key')
        logger.info('Keyvault: %s, key:%s', keyvault, '*' * len(secret_key))
        secret_data = get_keyvault_secret(keyvault, secret_key, vaulttoken)
        if secret_data and 'value' in secret_data:
            val = secret_data['value']
    return val


def set_azure_vault_data(secret_key=None, value=None):
    """Fetches the bearer token for Azure Vault API calls"""
    val = None
    vaulttoken = _get_vault_token()
    logger.debug('Vault Token: %s', vaulttoken)
    if vaulttoken and secret_key and value:
        keyvault = config_value('VAULT', 'keyvault')
        # secret_key = config_value('VAULT', 'secret_key')
        logger.info('Keyvault: %s, key:%s', keyvault, '*' * len(secret_key))
        sucess = set_keyvault_secret(keyvault, vaulttoken, secret_key, value)
        if sucess:
            return True
    return False

def set_azure_vault_data_with_response(secret_key=None, value=None):
    """Fetches the bearer token for Azure Vault API calls"""
    status, data = None, None
    vaulttoken = _get_vault_token()
    logger.debug('Vault Token: %s', vaulttoken)
    if vaulttoken and secret_key and value:
        keyvault = config_value('VAULT', 'keyvault')
        logger.info('Keyvault: %s, key:%s', keyvault, '*' * len(secret_key))
        status, data = set_keyvault_secret_with_response(keyvault, vaulttoken, secret_key, value)
    return status, data

def delete_azure_vault_data(secret_key=None):
    """"Delete a key from vault"""
    success = None
    vaulttoken = _get_vault_token()
    logger.debug('Vault Token: %s', vaulttoken)
    if vaulttoken and secret_key:
        keyvault = config_value('VAULT', 'keyvault')
        logger.info('Keyvault: %s, key:%s', keyvault, '*' * len(secret_key))
        success = delete_keyvault_secret(keyvault, secret_key, vaulttoken)
    logger.info('Secret Deleted: %s', success)
    return success


def get_cyberark_data(secret_key=None):
    """Get secret value for the secret key"""
    val = None
    ca_object = config_value('VAULT', 'CA_OBJECT')
    ca_safe = config_value('VAULT', 'CA_SAFE')
    ca_exe = config_value('VAULT', 'CA_EXE')
    ca_appid = config_value('VAULT', 'CA_APPID')
    if ca_object and ca_exe and ca_appid:
        cmd_args = '%s  GetPassword -p AppDescs.AppID=%s -p Query="Safe=%s;Folder=Root;Object=%s-%s" -o Password' \
                  % (ca_exe, ca_appid, ca_safe, ca_object, secret_key)
        my_process = Popen(cmd_args, shell=True, stdout=PIPE,
                                     stderr=PIPE,
                                     stdin=PIPE)
        out, err = my_process.communicate()
        err_result = err.rstrip() if err else None
        val = out.decode() if isinstance(out, bytes) else out.rstrip()
        if err_result:
            val = None
        else:
            logger.info('Secret Value: %s', '*' * len(val))
    return val



import json
import re
import collections
import traceback
import math

from processor.logging.log_handler import getlogger
logger = getlogger()


def shannon_entropy(data):
    if not data:
        return 0
    entropy = 0
    normalized_ent = 0
    n = 0
    for x in range(256):
        p_x = float(data.count(chr(x)))/len(data)
        if p_x > 0:
            n += 1
            entropy += - p_x*math.log(p_x, 2)

    if math.log(n) > 0:
        normalized_ent = entropy / math.log(n, 2)
    return entropy, normalized_ent


def get_paths(source):
    paths = []
    if isinstance(source, collections.MutableMapping):
        for k, v in source.items():
            paths.append([k])
            paths += [[k] + x for x in get_paths(v)]
    elif isinstance(source, collections.Sequence) and not isinstance(source, str):
        for i, v in enumerate(source):
            paths.append([i])
            paths += [[i] + x for x in get_paths(v)]
    return paths


def secret_finder(snapshot, PASSWORD_VALUE_RE, PASSWORD_KEY_RE=None, EXCLUDE_RE=None, shannon_entropy_password=False):
    output = {}
    errors = []
    try:
        issue_found = False
        skipped = True
        if isinstance(snapshot.get("resources"), list):
            for resource in snapshot.get("resources"):
                skipped = False
                path_list = get_paths(resource)
                for path in path_list:
                    nested_resource = resource
                    for key in path:
                        nested_resource = nested_resource[key]

                        if isinstance(nested_resource, (int, float, complex, bool)):
                            nested_resource = str(nested_resource)

                        if isinstance(nested_resource, str) and re.match(PASSWORD_VALUE_RE, nested_resource) and (re.match(PASSWORD_KEY_RE, str(key), re.I) if PASSWORD_KEY_RE else True) and (not(re.match(EXCLUDE_RE, str(nested_resource))) if EXCLUDE_RE else True):
                            if shannon_entropy_password:
                                _, normalized_entropy = shannon_entropy(
                                    nested_resource)
                                if normalized_entropy > 0.965:
                                    errors.append({
                                        "leaked_password_path": "resources/"+resource.get("type")+"/" + "/".join([str(path) for path in path]),
                                    })
                                    issue_found = True
                                    logger.warning("Leaked Password at:%s\nvalue:%s" % (
                                        "resources/"+resource.get("type")+"/" + "/".join([str(path) for path in path]), nested_resource))
                            else:
                                issue_found = True
                                errors.append({
                                    "leaked_password_path": "resources/"+resource.get("type")+"/" + "/".join([str(path) for path in path]),
                                })
                                logger.warning("Leaked Password at:%s\nvalue:%s" % (
                                    "resources/"+resource.get("type")+"/" + "/".join([str(path) for path in path]), nested_resource))

        output["issue"] = True if issue_found else False

        if errors:
            output["errors"] = errors

        output["skipped"] = skipped
        return output
    except Exception as ex:
        print(traceback.format_exc())
        output["issue"] = None
        output["err"] = str(ex)
        output["skipped"] = skipped
        return output


def azure_password_leak(generated_snapshot: dict) -> dict:

    PASSWORD_KEY_RE = r".*(?i)(password|securevalue|secret|privatekey|primarykey|secondarykey).*"
    PASSWORD_VALUE_RE = r'^(?!.*\$\{.*\}.*)(?=(?=.*[a-z][A-Z])|(?=.*[A-Z][a-z])|(?=.*[a-z][0-9])|(?=.*[0-9][a-z])|(?=.*[0-9][A-Z])|(?=.*[A-Z][0-9]))(.*[\^$*.\[\]{}\(\)?\-"!@\#%&\/,><\’:;|_~`]?)\S{8,99}$'
    EXCLUDE_REGEX = r'(?=.*([\[{(<]){1,})((\(.*\)){0,})((\[.*\]){0,})((\{.*\}){0,})((\<.*\>){0,})'

    output = secret_finder(
        generated_snapshot, PASSWORD_VALUE_RE, PASSWORD_KEY_RE, EXCLUDE_RE=EXCLUDE_REGEX)

    if output["issue"] == True:
        output["azure_password_leak"] = "Ensure no hardcoded password set in the template"

    elif output["issue"] == None:
        output["azure_password_leak"] = output["err"]
        output.pop("err")

    elif output["issue"] == False:
        output["azure_password_leak"] = ""
    return output


def entropy_password(generated_snapshot: dict) -> dict:

    PASSWORD_VALUE_RE = r'^(?!.*\$\{.*\}.*)(?=(?=.*[a-z][A-Z])|(?=.*[A-Z][a-z])|(?=.*[a-z][0-9])|(?=.*[0-9][a-z])|(?=.*[0-9][A-Z])|(?=.*[A-Z][0-9]))(?=.*[^A-Za-z0-9])\S{8,99}$'
    EXCLUDE_CONTAINS = ['API', 'AAD', 'Add', 'Advisor', 'AKS', 'Analysis', 'Analytics', 'Analyzer', 'API', 'App', 'Authorization', 'Automation', 'Azure', 'Batch', 'BI', 'Billing', 'Blockchain', 'Blueprints', 'Bot', 'Bus', 'Cache', 'CDN', 'Central', 'Certificate', 'Change', 'Cloud', 'Cognitive', 'Communication', 'Compute', 'Configuration', 'Consumption', 'Container', 'Cosmos', 'Custom', 'Customer', 'Data', 'Databricks', 'DB', 'Dedicated', 'Deployment', 'Device', 'DevOps', 'DevTest', 'Digital', 'DNS', 'Domain', 'Door', 'Event', 'Fabric', 'Factory', 'FarmBeats', 'for', 'Front', 'Graph', 'Grid', 'Hat', 'HDInsight', 'HSMs/', 'Hub', 'Hubs', 'Identity', 'Insights', 'Instance', 'IoT', 'Key', 'Kusto',
                        'Labs', 'Lake', 'Learning', 'Logic', 'Machine', 'Maintenance', 'Managed', 'Management', 'Manager', 'Maps', 'MariaDB', 'Media', 'Migrate', 'Migration', 'MySQL', 'NetApp', 'Network', 'Notification', 'Ons', 'OpenShift', 'Operational', 'Operations', 'Peering', 'Policy', 'Portal', 'PostgreSQL', 'Power', 'Providers', 'Provisioning', 'Recovery', 'Red', 'Registration', 'Registry', 'Relay', 'Resource', 'Resources', 'Scheduler', 'Search', 'Security', 'Series', 'Service', 'Services', 'Share', 'SignalR', 'Spring', 'SQL', 'Stack', 'Storage', 'Store', 'StorSimple', 'Stream', 'Subscription', 'Synapse', 'Sync', 'Time', 'Traffic', 'Twins', 'Update', 'Vault', 'Vaults', 'versions', 'Video', 'Virtual', 'Web', 'sha', 'sha256']
    EXCLUDE_REGEX = [
        "(?=^([a-zA-Z0-9]+\.+[a-zA-Z0-9]+\.[a-zA-Z0-9]+)(?![A-Za-z0-9])$)",
        "(?=^([a-zA-Z0-9]+\.+[a-zA-Z0-9]+\.[a-zA-Z0-9]+\.[a-zA-Z0-9]+)(?![A-Za-z0-9])$)",
        "(?=^([a-zA-Z0-9]+\_+[a-zA-Z0-9]+\_[a-zA-Z0-9]+)(?![A-Za-z0-9])$)",
        "(?=^([a-zA-Z0-9]+\-+[a-zA-Z0-9]+\-[a-zA-Z0-9]+)(?![A-Za-z0-9])$)",
        "(?=^([a-zA-Z0-9]+\.+[a-zA-Z0-9]*)$)",
        "(?=^(\/+\w{0,}){0,}$)",
        "(?=^([a-zA-Z0-9]*_[a-zA-Z0-9]*)$)",
        "(?=^([a-zA-Z0-9]*-[a-zA-Z0-9]*)$)",
    ]
    EXCLUDE_STARTSWITH = [
        "arn:"
    ]

    exclude_contains_regex = "(?=.*(?i)(%s).*)" % ("|".join(EXCLUDE_CONTAINS))

    exclude_startswith_regex = "(?=(%s))" % (
        "|".join(["^"+exclude for exclude in EXCLUDE_STARTSWITH]))
    exclude_regex = "|".join(EXCLUDE_REGEX)

    combined_exclude_regex = "|".join(
        [exclude_contains_regex, exclude_startswith_regex, exclude_regex])
    combined_exclude_regex = re.compile(combined_exclude_regex)

    output = secret_finder(
        generated_snapshot, PASSWORD_VALUE_RE, PASSWORD_KEY_RE=None, EXCLUDE_RE=combined_exclude_regex, shannon_entropy_password=True)

    if output["issue"] == True:
        output["entropy_password_err"] = "There is a possibility that a value might contains a secret string or password"

    elif output["issue"] == None:
        output["entropy_password_err"] = output["err"]
        output.pop("err")

    else:
        output["entropy_password_err"] = ""
    return output



"""Common file for running validator functions."""
# 1) github
# 1a) private
# HTTP - https://github.com/ajeybk/mytest.git
# username - kbajey@gmail.com or ajeybk
# password - prompt and env variable or form in the https url
# SSH - git@github.com:ajeybk/mytest.git
# identity file for host - /home/ajey/ssh_working/id_rsa_bitbucket

# 1b) public
# HTTP - https://github.com/prancer-io/cloud-validation-framework.git
# username - kbajey@gmail.com or ajeybk
# password - should not prompt, even if passed should work in url or from env variable, should work
# SSH - git@github.com:prancer-io/cloud-validation-framework.git
# identity file for host - /home/ajey/ssh_working/id_rsa_bitbucket

# 2) gitlab
# 2a) private
# HTTP - https://gitlab.com/tsdeepak/ulic_backend.git
# username - kbajey@gmail.com or kbajey
# password - prompt and env variable or form in the https url
# SSH - git@gitlab.com:kbajey/myrepo.git
# identity file for host - /home/ajey/ssh_working/id_rsa_bitbucket

# 2b) public
# HTTP - https://gitlab.com/kbajey/myrepo.git
# username - kbajey@gmail.com or kbajey
# password - prompt and env variable or form in the https url
# SSH - git@gitlab.com:kbajey/myrepo.git
# identity file for host - /home/ajey/ssh_working/id_rsa_bitbucket


# 3) bitbucket
# 3a) private
# HTTP - https://ajeybk@bitbucket.org/ajeybk/azframework.git
# username - kbajey@gmail.com or ajeybk
# password - prompt and env variable or form in the https url
# SSH - git@bitbucket.org:ajeybk/azframework.git
# identity file for host - /home/ajey/ssh_working/id_rsa_bitbucket

# 3b) public
# HTTP - https://ajeybk@bitbucket.org/ajeybk/aws-cli.git
# HTTP - https://ajeybk@bitbucket.org/ajeybk/mytestpub.git
# username - kbajey@gmail.com or ajeybk
# password - prompt and env variable or form in the https url
# SSH - git@bitbucket.org:ajeybk/aws-cli.git
# SSH - git@bitbucket.org:ajeybk/mytestpub.git
# identity file for host - /home/ajey/ssh_working/id_rsa_bitbucket

# 4) visualstudio
# 4a) private
# HTTP - https://ebizframework.visualstudio.com/whitekite/_git/whitekite
# HTTP - https://ajey.khanapuri%40liquware.com@ebizframework.visualstudio.com/whitekite/_git/whitekite
# username - ajey.khanapuri@liquware.com
# password - prompt and env variable or form in the https url
# SSH - Ebizframework@vs-ssh.visualstudio.com:v3/Ebizframework/whitekite/whitekite
# identity file for host - /home/ajey/ssh_working/id_rsa_azure
# 4b) public



#### Algorithm
# Add a attribute 'sshhost' : 'vs-ssh.visualstudio.com' or 'bitbucket.org' or 'gitlab.com' or 'github.com'
# Add a attribute 'sshuser': 'git' # All git servers expect 'git' as the user, but if there is an exception
# Add a attribute 'private': true|false
# For backward compatability it is assumed to be true.
# If giturl starts with https://, it is https based access otherwise ssh based access.

# For https public repo, username and password do not harm.
# For private https repo, read username from connector, if not present ignore.
# For https private repo, if username given, then read user_secret from connector, if present use,
# then check env variable, then prompt only

# For public ssh repo, ssh_key_file has to be present.
# Public repo clone with ssh_key_file does not require ssh/config file, whereas StrictHostKeyChecking=no
# may be required so that the user prompt may come up if githost is not present in known hosts.
# For private ssh repo with ssh_key_file, ssh/config file has to be created using 'sshhost', 'sshuser',
#  'ssh_key_file'
# Host <sshhost>
#   HostName <sshhost>
#   User <sshuser>
#   IdentityFile <ssh_key_file>
#
# Host *
#   IdentitiesOnly yes
#   ServerAliveInterval 100
import json
import hashlib
import time
import tempfile
import shutil
import hcl
import re
import os
import stat
import glob
import copy
import pymongo
from subprocess import Popen, PIPE
import urllib.parse
from git import Repo
from git import Git
from processor.helper.file.file_utils import exists_file, exists_dir, mkdir_path, remove_file
from processor.logging.log_handler import getlogger
from processor.helper.json.json_utils import get_field_value, json_from_file,\
    collectiontypes, STRUCTURE, get_field_value_with_default, \
    make_snapshots_dir, store_snapshot
from processor.helper.yaml.yaml_utils import yaml_from_file
from processor.helper.config.config_utils import config_value, get_test_json_dir, EXCLUSION
from processor.helper.config.rundata_utils import get_from_currentdata, get_dbtests, put_in_currentdata
from processor.database.database import insert_one_document, sort_field, get_documents,\
    COLLECTION, DATABASE, DBNAME, get_collection_size, create_indexes
from processor.helper.httpapi.restapi_azure import json_source
from processor.connector.snapshot_utils import validate_snapshot_nodes
from processor.connector.vault import get_vault_data
from processor.template_processor.base.base_template_constatns import TEMPLATE_NODE_TYPES
from processor.connector.git_connector.git_processor import git_clone_dir

logger = getlogger()

def convert_to_json(file_path, node_type):
    contentType = 'json'
    json_data = {}
    if node_type == 'json':
        json_data = json_from_file(file_path, escape_chars=['$'])
        contentType = 'json'
    elif node_type == 'terraform':
        contentType = 'terraform'
        with open(file_path, 'r') as fp:
            json_data = hcl.load(fp)
    elif node_type == 'yaml' or node_type == 'yml':
        contentType = 'yaml'
        json_data = yaml_from_file(file_path)
    else:
        logger.error("Snapshot error type:%s and file: %s", node_type, file_path)
    return contentType, json_data


def get_custom_data(snapshot_source, tabs=2):
    sub_data = {}
    if json_source():
        container = get_from_currentdata('container')
        dbname = config_value(DATABASE, DBNAME)
        collection = config_value(DATABASE, collectiontypes[STRUCTURE])
        parts = snapshot_source.split('.')
        qry = {'name': parts[0], 'container' : container }
        sort = [sort_field('timestamp', False)]
        docs = get_documents(collection, dbname=dbname, sort=sort, query=qry, limit=1)
        logger.info('Number of Custom Documents: %d', len(docs))
        if docs and len(docs):
            sub_data = docs[0]['json']
    else:
        json_test_dir = get_test_json_dir()
        file_name = '%s.json' % snapshot_source if snapshot_source and not \
            snapshot_source.endswith('.json') else snapshot_source
        custom_source = '%s/../%s' % (json_test_dir, file_name)
        logger.info('\t\tCUSTOM CONNECTOR: %s ', custom_source)
        # logger.info('Custom source: %s', custom_source)
        if exists_file(custom_source):
            sub_data = json_from_file(custom_source)
    return sub_data


def get_node(repopath, node, snapshot, ref, connector):
    """ Fetch node from the cloned git repository."""
    collection = node['collection'] if 'collection' in node else COLLECTION
    given_type = get_field_value(connector, "type")
    base_path = get_field_value_with_default(connector, "folderPath", "")
    snapshot_source = get_field_value(snapshot, 'source')
    parts = snapshot_source.split('.')
    db_record = {
        "structure": given_type,
        "reference": ref if not base_path else "",
        "source": parts[0],
        "path": base_path + node['path'],
        "timestamp": int(time.time() * 1000),
        "queryuser": get_field_value(snapshot, 'testUser'),
        "checksum": hashlib.md5("{}".encode('utf-8')).hexdigest(),
        "node": node,
        "snapshotId": node['snapshotId'],
        "collection": collection.replace('.', '').lower(),
        "json": {}
    }
    json_path = '%s/%s' % (repopath, node['path'])
    file_path = json_path.replace('//', '/')
    logger.info('\t\t\tFile: %s', file_path)
    if exists_file(file_path):
        node_type = node['type'] if 'type' in node and node['type'] else 'json'
        contentType, json_data = convert_to_json(file_path, node_type)
        # logger.info('type: %s, json:%s', node_type, json_data)
        # json_data = json_from_file(file_path)
        if json_data:
            db_record['contentType'] = contentType
            db_record['json'] = json_data
            data_str = json.dumps(json_data)
            db_record['checksum'] = hashlib.md5(data_str.encode('utf-8')).hexdigest()
    else:
        logger.info('Get requires valid file for snapshot not present!')
    logger.debug('DB: %s', db_record)
    return db_record

def get_all_nodes(repopath, node, snapshot, ref, connector):
    """ Fetch all the nodes from the cloned git repository in the given path."""
    db_records = []
    collection = node['collection'] if 'collection' in node else COLLECTION
    given_type = get_field_value(connector, "type")
    base_path = get_field_value_with_default(connector, "folderPath", "")
    snapshot_source = get_field_value(snapshot, 'source')
    parts = snapshot_source.split('.')
    d_record = {
        "structure": given_type,
        "reference": ref if not base_path else "",
        "source": parts[0],
        "path": '',
        "timestamp": int(time.time() * 1000),
        "queryuser": get_field_value(snapshot, 'testUser'),
        "checksum": hashlib.md5("{}".encode('utf-8')).hexdigest(),
        "node": node,
        "snapshotId": None,
        "masterSnapshotId": node['masterSnapshotId'],
        "collection": collection.replace('.', '').lower(),
        "json": {}
    }
    node_type = node['type'] if 'type' in node and node['type'] else 'json'
    json_path = '%s/%s' % (repopath, node['path'])
    file_path = json_path.replace('//', '/')
    logger.info('Dir: %s', file_path)
    if exists_dir(file_path):
        count = 0
        for filename in glob.glob('%s/*.json' % file_path.replace('//', '/')):
            parts = filename.rsplit('/', 1)
            path = '%s/%s' % (node['path'], parts[-1])
            contentType, json_data = convert_to_json(filename, node_type)
            logger.info('type: %s, json:%s', node_type, json_data)
            if json_data:
                db_record = copy.deepcopy(d_record)
                db_record['snapshotId'] = '%s%s' % (node['masterSnapshotId'], str(count))
                db_record['path'] = path.replace('//', '/')
                db_record['contentType'] = contentType
                db_record['json'] = json_data
                data_str = json.dumps(json_data)
                db_record['checksum'] = hashlib.md5(data_str.encode('utf-8')).hexdigest()
                db_records.append(db_record)
                count += 1
    else:
        logger.info('Get requires valid directory for snapshot not present!')
    return db_records

def _local_file_directory(connector, snapshot):
    final_path, repopath = None, None
    connector_user = get_field_value(connector, 'username')
    snapshot_user = get_field_value(snapshot, 'testUser')
    if snapshot_user == connector_user:
        folder_path = get_field_value(connector, 'folderPath')
        logger.info("Folder path: %s", folder_path)
        if exists_dir(folder_path):
            final_path = folder_path
        else:
            logger.error("Given folder path is not a directory")
        return repopath, final_path
    else:
        logger.error("Connector and snapshot user do not match.")
        return repopath, final_path 


def _get_repo_path(connector, snapshot):
    if connector and isinstance(connector, dict):
        git_provider = get_field_value(connector, "gitProvider")
        folder_path = get_field_value(connector, "folderPath")
        if git_provider:
            return git_clone_dir(connector)
        elif folder_path:
            return _local_file_directory(connector, snapshot)
    logger.error("Invalid connector or missing folderPath/gitProvider")
    return None, None


def populate_custom_snapshot(snapshot, container=None):
    """ Populates the resources from git."""
    if not container:
        container = get_from_currentdata('container')
    dbname = config_value('MONGODB', 'dbname')
    snapshot_source = get_field_value(snapshot, 'source')
    # connector_data = get_from_currentdata('connector')
    # if connector_data:
    #     sub_data = get_custom_data(connector_data)
    #     if not sub_data:
    #         logger.error("No connector data found in '%s'", connector_data)
    # else:
    #     sub_data = get_custom_data(snapshot_source)
    # Removed connector argument from prancer command line.
    sub_data = get_custom_data(snapshot_source)
    snapshot_nodes = get_field_value(snapshot, 'nodes')
    snapshot_data, valid_snapshotids = validate_snapshot_nodes(snapshot_nodes)
    if valid_snapshotids and sub_data and snapshot_nodes:
        baserepo, repopath = _get_repo_path(sub_data, snapshot)
        if repopath:
            brnch = get_field_value_with_default(sub_data, 'branchName', 'master')
            for node in snapshot_nodes:
                node_type = node['type'] if 'type' in node and node['type'] else ''
                if node_type in TEMPLATE_NODE_TYPES:
                    template_data = {
                        "container" : container,
                        "dbname" : dbname,
                        "snapshot_source" : snapshot_source,
                        "connector_data" : sub_data,
                        "snapshot_data" : snapshot_data,
                        "repopath" : repopath,
                        "snapshot" : snapshot
                    }
                    template_processor = TEMPLATE_NODE_TYPES[node_type](node, **template_data)
                    if 'snapshotId' in node:
                        snapshot_data = template_processor.populate_template_snapshot()
                    elif 'masterSnapshotId' in node:
                        snapshot_data = template_processor.populate_all_template_snapshot()
                elif 'paths' in node:
                    logger.error("ERROR: Invalid json : `%s` is not a valid node type." % node_type)
                else:
                    # logger.debug(node)
                    # data = get_node(repopath, node, snapshot_source, brnch)
                    # if data:
                    #     insert_one_document(data, data['collection'], dbname)
                    #     snapshot_data[node['snapshotId']] = True
                    validate = node['validate'] if 'validate' in node else True
                    if 'snapshotId' in node:
                        logger.debug(node)
                        data = get_node(repopath, node, snapshot, brnch, sub_data)
                        if data:
                            if validate:
                                if get_dbtests():
                                    if get_collection_size(data['collection']) == 0:
                                        #Creating indexes for collection
                                        create_indexes(data['collection'],
                                            config_value(DATABASE, DBNAME), 
                                            [('snapshotId', pymongo.ASCENDING),
                                            ('timestamp', pymongo.DESCENDING)])
                                            
                                        create_indexes(
                                            data['collection'], 
                                            config_value(DATABASE, DBNAME), 
                                            [
                                                ('_id', pymongo.DESCENDING),
                                                ('timestamp', pymongo.DESCENDING),
                                                ('snapshotId', pymongo.ASCENDING)
                                            ]
                                        )
                                    insert_one_document(data, data['collection'], dbname)
                                else:
                                    snapshot_dir = make_snapshots_dir(container)
                                    if snapshot_dir:
                                        store_snapshot(snapshot_dir, data)
                                if 'masterSnapshotId' in node:
                                    snapshot_data[node['snapshotId']] = node['masterSnapshotId']
                                else:
                                    snapshot_data[node['snapshotId']] = True
                            # else:
                            #     snapshot_data[node['snapshotId']] = False
                            node['status'] = 'active'
                        else:
                            node['status'] = 'inactive'
                        logger.debug('Type: %s', type(data))
                    elif 'masterSnapshotId' in node:
                        includeSnapshotConfig = get_from_currentdata("INCLUDESNAPSHOTS")
                        includeSnapshots = get_from_currentdata("SNAPHSHOTIDS")
                        ignoreExclusion = False
                        ignoreNode = False
                        if includeSnapshotConfig:
                            if node['masterSnapshotId'] in includeSnapshots:
                                ignoreExclusion = True
                                ignoreNode = False
                            else:
                                ignoreNode  = True
                        alldata = get_all_nodes(repopath, node, snapshot, brnch, sub_data)
                        if alldata:
                            exclusions = get_from_currentdata(EXCLUSION).get('exclusions', [])

                            resourceExclusions = {}
                            if not ignoreExclusion:
                                for exclusion in exclusions:
                                    if 'exclusionType' in exclusion and exclusion['exclusionType'] and exclusion['exclusionType'] == 'resource':
                                        if 'paths' in exclusion and isinstance(exclusion['paths'], list):
                                            resourceExclusions[tuple(exclusion['paths'])] = exclusion

                            snapshot_data[node['masterSnapshotId']] = []
                            for data in alldata:
                                if isinstance(data['path'], str):
                                    key = tuple([data['path']])
                                elif isinstance(data['path'], list):
                                    key = tuple(data['path'])
                                else:
                                    key = None
                                if key and key in resourceExclusions:
                                    logger.warning("Excluded from resource exclusions: %s", data['path'])
                                    continue
                                if not ignoreNode:
                                    snapshot_data[node['masterSnapshotId']].append(
                                        {
                                            'snapshotId': data['snapshotId'],
                                            'path': data['path'],
                                            'validate': validate
                                        })
                        logger.debug('Type: %s', type(alldata))
        if baserepo and os.path.exists(baserepo):
            # logger.info('\t\tCLEANING Repo: %s', baserepo)
            put_in_currentdata("CLEANING_REPOS", baserepo)
            # shutil.rmtree(baserepo)
    return snapshot_data


def main():
    connectors = [
        {
            "fileType": "structure",
            "type": "filesystem",
            "companyName": "prancer-test",
            "gitProvider": "https://github.com/ajeybk/mytest.git",
            "branchName": "master",
            "username": None,
            "password": None,
            "sshKeyfile": None,
            "private": True,
            "sshHost": "github.com",
            "sshUser": "git"
        },
        {
            "fileType": "structure",
            "type": "filesystem",
            "companyName": "prancer-test",
            "gitProvider": "https://github.com/ajeybk/mytest.git",
            "branchName": "master",
            "username": "kbajey@gmail.com",
            "password": None,
            "sshKeyfile": None,
            "private": True,
            "sshHost": "github.com",
            "sshUser": "git"
        },
        {
            "fileType": "structure",
            "type": "filesystem",
            "companyName": "prancer-test",
            "gitProvider": "https://github.com/ajeybk/mytest.git",
            "branchName": "master",
            "username": "ajeybk",
            "password": None,
            "sshKeyfile": None,
            "private": True,
            "sshHost": "github.com",
            "sshUser": "git"
        },
        {
            "fileType": "structure",
            "type": "filesystem",
            "companyName": "prancer-test",
            "gitProvider": "git@github.com:ajeybk/mytest.git",
            "branchName": "master",
            "username": "ajeybk",
            "sshKeyfile": "/home/ajey/ssh_working/id_rsa_bitbucket",
            "private": True,
            "sshHost": "github.com",
            "sshUser": "git"
        },
        {
            "fileType": "structure",
            "type": "filesystem",            
            "companyName": "prancer-test",
            "gitProvider": "https://github.com/prancer-io/cloud-validation-framework.git",
            "branchName": "master",
            "username": "kbajey@gmail.com",
            "password": None,
            "sshKeyfile": None,
            "private": False,
            "sshHost": "github.com",
            "sshUser": "git"
        },
        {
            "fileType": "structure",
            "type": "filesystem",
            "companyName": "prancer-test",
            "gitProvider": "git@github.com:prancer-io/prancer-hello-world.git",
            "branchName": "master",
            "sshKeyfile": None,
            "sshKeyName": "<thenameofthekeyinthevault>",
            "sshUser": "git",
            "sshHost": "github.com",
            "private": True
        }
    ]
    for conn in connectors:
        logger.info('#' * 50)
        repodir, clonedir = git_clone_dir(conn)
        logger.info("Delete: %s, clonedir: %s", repodir, clonedir)



if __name__ == "__main__":
    main()



"""Helper functions to setup logging for the framework."""
import logging
from logging.handlers import RotatingFileHandler
import datetime
import time
import os
import json
from pymongo import MongoClient
from pymongo.errors import ServerSelectionTimeoutError

FWLOGGER = None
FWLOGFILENAME = None
MONGOLOGGER = None
DBLOGGER = None
dbhandler = None
DEFAULT_LOGGER = None
LOGLEVEL = None
DEBUG_LOGFORMAT = '%(asctime)s(%(module)s:%(lineno)4d) - %(message)s'
LOGFORMAT = '%(asctime)s - %(message)s'


def get_dblog_name():
    """ Set as per current datetime formay, could be passed thru an environment variable"""
    dblog_name = os.getenv('DBLOG_NAME', None)
    if not dblog_name:
        dblog_name = 'logs_%s' % datetime.datetime.now().strftime('%Y%m%d%H%M%S')
    return dblog_name

def default_logger():
    global DEFAULT_LOGGER
    if DEFAULT_LOGGER:
        return DEFAULT_LOGGER
    logging.basicConfig(format=LOGFORMAT)
    DEFAULT_LOGGER = logging.Logger(__name__)
    
    handler = DefaultLoggingHandler()
    handler.setFormatter(ColorFormatter(LOGFORMAT))
    DEFAULT_LOGGER.addHandler(handler)
    return DEFAULT_LOGGER

def get_logformat(log_level):
    if log_level == "DEBUG":
        return DEBUG_LOGFORMAT
    return LOGFORMAT

def get_loglevel(fwconf=None):
    """ Highest priority is at command line, then ini file otherwise default is INFO"""
    global LOGLEVEL
    if LOGLEVEL:
        return LOGLEVEL

    default_logger()
    loglevels = ['CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG']
    level = os.getenv('LOGLEVEL', None)
    loglevel = None
    if level and level in loglevels:
        loglevel = level
    elif level and level not in loglevels:
        DEFAULT_LOGGER.warning("Invalid log level passed in parameter \"%s\", valid log levels are %s" % (level, ", ".join(loglevels)))
        os.environ["LOGLEVEL"] = "INFO"

    cfglevel =  None
    if fwconf and 'level' in fwconf and fwconf['level'] and fwconf['level'].upper() in loglevels:
        cfglevel = fwconf['level'].upper()
    elif fwconf and 'level' in fwconf and fwconf['level'] and fwconf['level'].upper() not in loglevels:
        DEFAULT_LOGGER.warning("Invalid log level set in config file \"%s\", valid log levels are %s" % (fwconf['level'], ", ".join(loglevels)))

    if loglevel:
        LOGLEVEL = loglevel
    elif cfglevel:
        LOGLEVEL = cfglevel
    else:
        return logging.getLevelName(logging.INFO)
    return LOGLEVEL

class DefaultLoggingHandler(logging.StreamHandler):

    def emit(self, record):
        """
        Emit a record.

        If a formatter is specified, it is used to format the record.
        The record is then written to the stream with a trailing newline.  If
        exception information is present, it is formatted using
        traceback.print_exception and appended to the stream.  If the stream
        has an 'encoding' attribute, it is used to determine how to do the
        output to the stream.
        """
        try:
            log_msg = self.format(record)
            stream = self.stream
            stream.write(log_msg)
            stream.write(self.terminator)
            self.flush()
        except UnicodeEncodeError:
            msg = self.format(record)
            stream = self.stream
            stream.write(str(msg.encode('utf-8')))
            stream.write(self.terminator)
            self.flush()
        except Exception:
            self.handleError(record)


class DefaultFileHandler(logging.FileHandler):

    def emit(self, record):
        """
        Emit a record.

        If the stream was not opened because 'delay' was specified in the
        constructor, open it before calling the superclass's emit.
        """
        if self.stream is None:
            self.stream = self._open()
        DefaultLoggingHandler.emit(self, record)

        # try:
        #     log_msg = self.format(record)
        #     db_record = {
        #         "timestamp": int(time.time() * 1000),
        #         "level": record.levelname,
        #         "module": record.module,
        #         "line": record.lineno,
        #         "asctime": record.asctime,
        #         "msg": log_msg,
        #         "log_type": getattr(record, "type", "DEFAULT")
        #     }
        #     msg = json.dumps(db_record)
        #     stream = self.stream
        #     stream.write(msg)
        #     stream.write(self.terminator)
        #     self.flush()
        # except UnicodeEncodeError:
        #     msg = self.format(record)
        #     stream = self.stream
        #     stream.write(str(msg.encode('utf-8')))
        #     stream.write(self.terminator)
        #     self.flush()
        # except Exception:
        #     self.handleError(record)


class DefaultRoutingFileHandler(RotatingFileHandler):

    def __init__(self, dbargs, filename, maxBytes=0, backupCount=0):
        self.isjson = True if dbargs == 3 else False
        RotatingFileHandler.__init__(self, filename, maxBytes=maxBytes, backupCount=backupCount)

    def emit(self, record):
        """
        Emit a record.

        Output the record to the file, catering for rollover as described
        in doRollover().
        """
        try:
            if self.shouldRollover(record):
                self.doRollover()
            if self.stream is None:
                self.stream = self._open()
            # DefaultFileHandler.emit(self, record)
            stream = self.stream
            if self.isjson:
                log_msg = self.format(record)
                db_record = {
                    "timestamp": int(time.time() * 1000),
                    "level": record.levelname,
                    "module": record.module,
                    "line": record.lineno,
                    "asctime": record.asctime,
                    "msg": log_msg,
                    "log_type": getattr(record, "type", "DEFAULT")
                }
                msg = json.dumps(db_record)
            else:
                msg = self.format(record)
            stream.write(msg)
            stream.write(self.terminator)
            self.flush()
        except UnicodeEncodeError:
            msg = self.format(record)
            stream = self.stream
            stream.write(str(msg.encode('utf-8')))
            stream.write(self.terminator)
            self.flush()
        except Exception:
            self.handleError(record)


class MongoDBHandler(logging.Handler):
    """Customized logging handler that puts logs to the database, pymongo required
    """
    def __init__(self, dburl, dbname, collection='logs'):
        logging.Handler.__init__(self)
        try:
            dbconnection =  MongoClient(host=dburl, serverSelectionTimeoutMS=3000)
            _ = dbconnection.list_database_names()
            if dbname:
                db = dbconnection[dbname]
            else:
                db = dbconnection['test']
            # Collection where the log statements are being put in.
            self.coll_name = collection
            # Every run of the prancer-basic will have a log name associated with it.
            self.log_name = ''
            if db:
                self.db = db
                self.set_log_collection()
            else:
                self.collection = None
        except ServerSelectionTimeoutError as ex:
            self.collection = None

    def set_log_collection(self):
        global DBLOGGER
        self.collection = self.db[self.coll_name]
        DBLOGGER = get_dblog_name()
        self.dblog_name = DBLOGGER
        self.collection.insert({'name': self.dblog_name, 'logs': []}, check_keys=False)

    def get_log_collection(self):
        return self.dblog_name

    def reset_log_collection(self):
        self.dblog_name = ''

    def emit(self, record):
        """Add record to the database"""
        # format the log message so it can be put to db (escape quotes)
        self.log_msg = self.format(record)
        db_record = {
            "timestamp": int(time.time() * 1000),
            "level": record.levelname,
            "module": record.module,
            "line": record.lineno,
            "asctime": record.asctime,
            "msg": self.log_msg,
            "log_type": getattr(record, "type", "DEFAULT")
        }

        try:
            if self.collection and self.dblog_name:
                # self.collection.insert(db_record, check_keys=False)
                self.collection.update({'name': self.dblog_name}, {'$push': {'logs': db_record}})
        except Exception as e:
            print('CRITICAL Logger DB ERROR: Logging to database not possible!')

class CustomLogger(logging.Logger):
    def makeRecord(self, name, level, fn, lno, msg, args, exc_info, func=None, extra=None, sinfo=None):
        log_record = logging.Logger.makeRecord(self, name, level, fn, lno, msg, args, exc_info, func, extra)
        if extra and isinstance(extra, dict):
            for key, value in extra.items():
                log_record.__dict__.setdefault(key, value)
        return log_record


class ColorFormatter(logging.Formatter):
    """Logging Formatter to add colors and count warning / errors"""

    grey = "\033[90m"
    yellow = "\033[93m"
    red = "\033[91m"
    bold_red = "\x1b[31;1m"
    success = "\033[92m"
    blue = "\033[94m"
    default = "\033[99m"
    reset = "\x1b[0m"

    def __init__(self, log_format):
        self.log_format = log_format
        self.COLOR_FORMATS = {
            # for set color based on record type from extra parameters
            "debug" : self.grey + log_format + self.reset,
            "info" : self.blue + log_format + self.reset,
            "success" : self.success + log_format + self.reset,
            "default": self.default + log_format + self.reset,
            "critical" : self.bold_red + log_format + self.reset,

            # for set color based on record levelname
            # "CRITICAL" : self.bold_red + log_format + self.reset,
            "WARNING" : self.yellow + log_format + self.reset,
            "ERROR" : self.red + log_format + self.reset
        }

    def format(self, record):
        levelname = record.levelname
        log_fmt = self.COLOR_FORMATS.get(levelname)
        if not log_fmt:
            log_type = getattr(record, "type", "DEFAULT")
            log_fmt = self.COLOR_FORMATS.get(log_type.lower())
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)

def get_logdir(fw_cfg, baselogdir):
    """ Given ini logging config and base framework dir, checks if the log folder is writeable."""
    if not fw_cfg:
        return False, None
    log_writeable = True
    if fw_cfg and 'LOGGING' in fw_cfg:
        fwconf = fw_cfg['LOGGING']
        if 'logFolder' in fwconf and fwconf['logFolder'] and os.path.isdir(baselogdir):
            logdir = '%s/%s' % (baselogdir, fwconf['logFolder'])
    try:
        if not os.path.exists(logdir):
            os.makedirs(logdir)
    except:
        log_writeable = False
    try:
        if log_writeable:
            from pathlib import Path
            testfile = '%s/%d' % (logdir, int(time.time()))
            Path(testfile).touch()
            if os.path.exists(testfile):
                os.remove(testfile)
            else:
                log_writeable = False
    except:
        log_writeable = False
    return log_writeable, logdir



def ini_logging_config(fwconfigfile):
    """logging config"""
    from processor.helper.config.config_utils import framework_config, get_config_data, framework_dir, get_base_log_dir
    if not fwconfigfile:
        fwconfigfile = framework_config()
    fw_cfg = get_config_data(fwconfigfile)
    log_config = {
        "level": logging.INFO,
        "propagate": True,
        "size": 10,
        "backups": 10,
        "db": None,
        'logpath': None
    }
    if fw_cfg and 'LOGGING' in fw_cfg:
        base_log_dir = get_base_log_dir()
        if base_log_dir is None:
            base_log_dir = framework_dir()
        logwriteable, logpath = get_logdir(fw_cfg, base_log_dir)
        if logwriteable and logpath:
            log_config['logpath'] = logpath
        fwconf = fw_cfg['LOGGING']
        log_config['level'] = get_loglevel(fwconf)
        log_config['size'] = fwconf.getint('size') if 'size' in fwconf else 10
        log_config['backups'] = fwconf.getint('backups') if 'backups' in fwconf else 10
        log_config['propagate'] = fwconf.getboolean('propagate') if 'propagate' in fwconf else True
        log_config['db'] = fwconf['dbname'] if 'dbname' in fwconf else None
    return log_config


def default_logging(fwconfigfile=None):
    """Framework default logging to console"""
    log_config = None
    if fwconfigfile:
        log_config = ini_logging_config(fwconfigfile)

    log_level = get_loglevel(log_config)
    log_format = get_logformat(log_level)

    # logging.basicConfig(format=LOGFORMAT)
    logger = CustomLogger(__name__)
    logger.propagate = log_config['propagate'] if log_config and 'propagate' in log_config else True
    logger.setLevel(get_loglevel(log_config))

    handler = DefaultLoggingHandler()
    handler.setFormatter(ColorFormatter(log_format))
    logger.addHandler(handler)
    return logger


def add_file_logging(fwconfigfile, dbargs):
    """ Add file logging to the basic logging"""
    global FWLOGGER, FWLOGFILENAME
    log_config = ini_logging_config(fwconfigfile)
    if not log_config['logpath']:
        return
    dblogname = os.getenv('DBLOG_NAME', None)
    logname  = dblogname if dblogname else datetime.datetime.today().strftime('%Y%m%d-%H%M%S')
    FWLOGFILENAME = '%s/%s.log' % (log_config['logpath'], logname)
    if not FWLOGGER:
        FWLOGGER = default_logging()
    handler = DefaultRoutingFileHandler(
        dbargs,
        FWLOGFILENAME,
        maxBytes=1024 * 1024 * log_config['size'],
        backupCount=log_config['backups']
    )
    handler.setFormatter(logging.Formatter(get_logformat(log_config['level'])))
    handler.setLevel(log_config['level'])
    FWLOGGER.addHandler(handler)


def add_db_logging(fwconfigfile, dburl, dbargs):
    """ Add database logging to the basic logging"""
    global FWLOGGER, dbhandler
    log_config = ini_logging_config(fwconfigfile)
    unittest = os.getenv('UNITTEST', "false")
    if log_config['db'] and unittest != "true" and dbargs in [1,2] and dburl:
        if not FWLOGGER:
            FWLOGGER = default_logging()
        dblogformat = get_logformat(log_config['level'])
        dbhandler = MongoDBHandler(dburl, log_config['db'])
        dbhandler.setFormatter(logging.Formatter(dblogformat))
        dbhandler.setLevel(log_config['level'])
        FWLOGGER.addHandler(dbhandler)


def getlogger(fw_cfg=None):
    """Get the logger for the framework."""
    global FWLOGGER
    if FWLOGGER:
        return FWLOGGER
    FWLOGGER = default_logging()
    return FWLOGGER

def logging_fw(fwconfigfile, dbargs, refresh_logger=False):
    """Framework file logging"""
    global FWLOGGER
    if FWLOGGER and (dbhandler and dbargs == 2) and not refresh_logger:
        return FWLOGGER
    FWLOGGER = default_logging(fwconfigfile)
    add_file_logging(fwconfigfile, dbargs)
    unittest = os.getenv('UNITTEST', "false")
    if unittest != "true":
        from processor.logging.dburl_kv import get_dburl
        dburl = get_dburl()
        add_db_logging(fwconfigfile, dburl, dbargs)
    return FWLOGGER


def init_logger(dbargs, fw_cfg=None, refresh_logger=False):
    """Get the logger for the framework."""
    return logging_fw(fw_cfg, dbargs, refresh_logger)


def init_logger_old(dbargs, fw_cfg=None, refresh_logger=False):
    """Get the logger for the framework."""
    global FWLOGGER
    if FWLOGGER and (dbhandler and dbargs == 'FULL') and not refresh_logger:
        return FWLOGGER
    FWLOGGER = logging_fw(fw_cfg, dbargs)
    return FWLOGGER


def get_dblogger():
    return DBLOGGER


def get_dblog_handler():
    return dbhandler




from jinja2 import Undefined
from jinja2._compat import implements_to_string, string_types
from jinja2.utils import missing, object_type_repr
from processor.logging.log_handler import getlogger

logger = getlogger()

class ResourceContext(object):

    def __init__(self, properties={}, **kwargs):
        self.properties = properties
    
    def __getattribute__(self, name):
        __dict__ = super(ResourceContext, self).__getattribute__('__dict__')
        value = {}
        if name in __dict__:
            value = object.__getattribute__(self, name)
        return value


@implements_to_string
class SilentUndefined(Undefined):
    '''
    handle undefined variables
    '''
    def _fail_with_undefined_error(self, *args, **kwargs):
        if self._undefined_hint is None:
            if self._undefined_obj is missing:
                hint = '%r is undefined' % self._undefined_name
            elif not isinstance(self._undefined_name, string_types):
                hint = '%s has no element %r' % (
                    self._undefined_obj,
                    self._undefined_name
                )
            else:
                hint = '%r has no attribute %r' % (
                    self._undefined_obj,
                    self._undefined_name
                )
        else:
            hint = self._undefined_hint
        # logger.error(hint)
        return ''

    __slots__ = ()
    __iter__ = __str__ = __len__ = __nonzero__ = __eq__ = \
        __ne__ = __bool__ = __hash__ = \
        _fail_with_undefined_error
    
    def __getattr__(self, name):
        if name[:2] == '__':
            raise AttributeError(name)
        return self._fail_with_undefined_error()

    __add__ = __radd__ = __mul__ = __rmul__ = __div__ = __rdiv__ = \
        __truediv__ = __rtruediv__ = __floordiv__ = __rfloordiv__ = \
        __mod__ = __rmod__ = __pos__ = __neg__ = __call__ = \
        __getitem__ = __lt__ = __le__ = __gt__ = __ge__ = __int__ = \
        __float__ = __complex__ = __pow__ = __rpow__ = __sub__ = \
        __rsub__ = _fail_with_undefined_error

    def __eq__(self, other):
        return type(self) is type(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return id(type(self))

    def __str__(self):
        return u''

    def __len__(self):
        return 0

    def __iter__(self):
        if 0:
            yield None

    def __nonzero__(self):
        return False
    __bool__ = __nonzero__

    def __repr__(self):
        return 'Undefined'


"""
Define an interface for parsing azure template and its parameter files.
"""
import copy
import glob
import re
import os
import inspect
import ast
import hcl2
from processor.logging.log_handler import getlogger
from processor.templates.base.template_parser import TemplateParser
from processor.helper.file.file_utils import exists_file, exists_dir
from processor.templates.terraform.helper.function.terraform_functions import default_functions
from processor.templates.terraform.helper.expression.terraform_expressions import expression_list
from processor.helper.json.json_utils import json_from_file, json_from_string
from processor.helper.hcl.hcl_utils import hcl_to_json
from processor.templates.terraform.helper.module_parser import ModuleParser

logger = getlogger()

class TerraformTemplateParser(TemplateParser):
    """
    Terraform Parser class for process terraform template
    """
    def __init__(self, template_file, tosave=False, **kwargs):
        super().__init__(template_file, tosave=False, **kwargs)
        self.imports = []
        self.schema_filter = {
            "var" : self.process_variable,
            "data" : self.process_data,
            "other" : self.process_other,
        }
        self.default_gparams = kwargs.get("default_gparams", {})
        self.gdata = {}
        self.resource = {}
        self.module_params = {
            "module" : {}
        }
        self.temp_params = {}
        self.count = None
        self.process_module = kwargs.get("process_module", False)
        self.replace_values = {
            "true" : True,
            "false" : False,
            "True" : True,
            "False" : False,
            "&&" : "and",
            "||" : "or",
            "None" : None,
            "!" : "not",
        }
        self.replace_value_str = ["and", "or", "not"]
        self.template_file_list = [self.template_file]
        self.parameter_file_list = self.parameter_file if self.parameter_file else []
        self.connector_data = kwargs.get("connector_data", False)
        self.exclude_directories = [".git"]
        self.template_references = {
            "main_templates" : [],
            "module_templates" : []
        }
        self.outputs = {}
        self.skip_key_to_process = ["compiletime_identity"]

    def is_template_file(self, file_path):
        """
        check for valid template file for parse terraform template
        """
        if len(file_path.split(".")) > 0 and file_path.split(".")[-1] == "tf":
            json_data = hcl_to_json(file_path)
            return True if (json_data and ("resource" in json_data or "module" in json_data)) else False
        elif len(file_path.split(".")) > 0 and file_path.split(".")[-1] == "json":
            json_data = json_from_file(file_path, escape_chars=['$'])
            return True if (json_data and ("resource" in json_data or "module" in json_data)) else False
    
    def is_parameter_file(self, file_path):
        """
        check for valid variable file for parse terraform template
        """
        if len(file_path.split(".")) > 0 and file_path.split(".")[-1] in ["tf", "tfvars"]:
            json_data = hcl_to_json(file_path)
            return True if (json_data and not "resource" in json_data) else False
        elif len(file_path.split(".")) > 1 and [ele for ele in [".tfvars.json", ".tf.json"] if(ele in file_path)]:
            json_data = json_from_file(file_path, escape_chars=['$'])
            return True if (json_data and not "resource" in json_data) else False
        return False

    def process_other(self, resource):
        """
        return default value of the variable if variable is set in terraform files
        """
        main_status = False

        status, value = self.parse_field_value(resource, self.resource)
        main_status = True if status else main_status
        if value is None:
            status, value = self.parse_field_value(resource, self.module_params)
            main_status = True if status else main_status
        if value is None:
            status, value = self.parse_field_value(resource, self.temp_params)
            main_status = True if status else main_status
        return main_status, value

    def process_variable(self, resource):
        """
        return default value of the variable if variable is set in terraform files
        """
        main_status = False
        status, value = self.parse_field_value(resource, self.default_gparams)
        main_status = True if status else main_status
        if value is None:
            status, value = self.parse_field_value(resource, self.gparams)
            main_status = True if status else main_status
        if value is None:
            status, value = self.parse_field_value(resource, self.resource)
            main_status = True if status else main_status
        return main_status, value
    
    def process_data(self, resource):
        """
        return value of the data if the given data is contains in terraform template file
        """
        return self.parse_field_value(resource, self.gdata)
    
    def parse_field_value(self, resource, from_data):
        resource_list = re.compile("\.(?![^[]*\])").split(resource)
        value = copy.deepcopy(from_data)
        is_valid = True
        for resource in resource_list:
            # process variables like this: var.network_http["cidr"]
            pattern = re.compile(r'\[["|\'](.*?)["|\']\]')
            map_keys = re.findall(pattern, resource)
            if not map_keys:
                pattern = re.compile(r'\[(.*?)\]')
                map_keys = re.findall(pattern, resource)
            if map_keys and len(resource.split("[")) > 1:
                res = resource.split("[")[0]
                if res in value:
                    value = value.get(res)
                else:
                    value = None
                    is_valid = False
                    break
                for key in map_keys:
                    if value is None or not isinstance(value, dict):
                        is_valid = False
                        break

                    if key in value:
                        value = value.get(key)
                    else:
                        value = None
                        is_valid = False
                if value is None:
                    break
                continue
            if isinstance(value, list) and resource == "*" and self.count is not None and self.count < len(value):
                value = value[self.count]
            elif isinstance(value, dict):            
                if resource in value:
                    value = value.get(resource)
                else:
                    value = None
                    is_valid = False
                if value is None:
                    break
            else:
                value = None
                is_valid = False
                break
        
        if is_valid:
            return True, value
        else:
            return False, value
    
    def get_template(self):
        """
        return the template json data
        """
        # used for terraform file remediation
        self.template_references["main_templates"].append({
            "main_file_path" : self.get_ralative_path(self.template_file)
        })

        json_data = None
        if len(self.template_file.split(".")) > 0 and self.template_file.split(".")[-1]=="tf":
            json_data = hcl_to_json(self.template_file)
            self.contentType = 'terraform'
        elif len(self.template_file.split(".")) > 1 and ".tf.json" in self.template_file:
            json_data = json_from_file(self.template_file, escape_chars=['$'])
        return json_data
    
    def get_paramter_json_list(self):
        """
        process parameter files and returns parameters json list
        """
        parameter_json_list = []
        for parameter in self.parameter_file:
            if len(parameter.split(".")) > 0 and parameter.split(".")[-1] in ["tf", "tfvars"]:
                json_data = hcl_to_json(parameter)
                if json_data:
                    parameter_json_list.append({parameter.split(".")[-1] : json_data})
            elif len(parameter.split(".")) > 1 and [ele for ele in [".tfvars.json", ".tf.json"] if(ele in parameter)]:
                json_data = json_from_file(parameter, escape_chars=['$'])
                if json_data:
                    splited_list = parameter.split(".")
                    parameter_json_list.append({'.'.join(splited_list[len(splited_list)-2:]) : json_data})
        return parameter_json_list
    
    def get_ralative_path(self, file_path):
        """
        takes full path of template or parameter file and returns the relative path by removing `temp` directory path
        """
        split_path = file_path.split("/")
        return "/%s" % "/".join(split_path[3:]).replace("//","/")

    def generate_template_json(self):
        """
        generate the template json from template and parameter files
        """
        template_json = self.get_template()
        parameter_jsons = self.get_paramter_json_list()
        gen_template_json = None
        if template_json:
            gen_template_json = copy.deepcopy(template_json)
            for parameter_json in parameter_jsons:
                for file_type, variable_json in parameter_json.items():
                    if file_type in ["tfvars", "tfvars.json"]:
                        for key, value in variable_json.items():
                            if isinstance(value, list) and len(value) == 1 and isinstance(value[0], str):
                                value[0] = self.parse_string(value[0])
                            key = self.parse_string(key)
                            self.gparams[key] = value
                    else:
                        if "variable" in variable_json:
                            for var in variable_json["variable"]:
                                for key, value in var.items():
                                    if "default" in value:
                                        self.gparams[key] = value["default"] 

            if "variable" in template_json:
                for var in template_json["variable"]:
                    for key, value in var.items():
                        if "default" in value:
                            self.gparams[key] = value["default"]

            new_resources = {}
            if "module" in template_json:
                for mod in template_json["module"]:
                    for key, value in mod.items():
                        if "source" in value:
                            default_gparams = {}
                            self.module_params["module"][key] = {}
                            for k, v in value.items():
                                if k != "source":
                                    processed_data, processed = self.process_resource(v, count=self.count)
                                    default_gparams[k] = processed_data
                                    self.module_params["module"][key][k] = processed_data

                            full_path_list = self.template_file.split("/")[:-1]
                            template_file_path = ("/".join(full_path_list)).replace("//","/")
                            module_parser = ModuleParser(value["source"], template_file_path, connector_data=self.connector_data)
                            module_file_path = module_parser.process_source()

                            if not module_file_path:
                                continue
                            
                            logger.info("Finding module : %s", module_file_path)
                            if exists_dir(module_file_path):
                                list_of_file = os.listdir(module_file_path)

                                template_file_path_list = []
                                parameter_file_list = []
                                for entry in list_of_file:
                                    if any(exclude_dir in entry for exclude_dir in self.exclude_directories):
                                        continue
                                    new_file_path = ('%s/%s' % (module_file_path, entry)).replace('//', '/')
                                    self.template_references["module_templates"].append({
                                        "main_file_path" : self.get_ralative_path(self.template_file),
                                        "module_file_path" : self.get_ralative_path(new_file_path),
                                        "module_label" : key
                                    })
                                    if exists_file(new_file_path):
                                        if self.is_template_file(new_file_path):
                                            template_file_path_list.append(new_file_path)
                                        elif self.is_parameter_file(new_file_path):
                                            parameter_file_list.append(new_file_path)
                                
                                if template_file_path_list and parameter_file_list:
                                    for template_file_path in template_file_path_list:
                                        terraform_template_parser = TerraformTemplateParser(
                                            template_file_path,
                                            parameter_file=parameter_file_list,
                                            **{"default_gparams" : default_gparams, "process_module" : True })
                                        new_template_json = terraform_template_parser.parse()

                                        self.template_file_list = self.template_file_list + terraform_template_parser.template_file_list
                                        self.parameter_file_list = self.parameter_file_list + terraform_template_parser.parameter_file_list

                                        # TO DO: Update module peram
                                        # for out_key, out_value in terraform_template_parser.outputs.items():
                                        #     self.module_params["module"][key][out_key] = out_value
                                        
                                        if new_template_json:
                                            for resource, resource_item in new_template_json.items():
                                                # set parameters from modules files to main resource file
                                                if resource == "resource":
                                                    for resource_key, resource_value in resource_item.items():
                                                        for resource_name, resource_properties in resource_value.items():
                                                            if isinstance(resource_properties, dict):
                                                                for default_key, default_value in default_gparams.items():
                                                                    if default_key not in resource_properties:
                                                                        resource_properties[default_key] = default_value
                                                                resource_properties["compiletime_identity"] = "module.%s" % key
                                                            if isinstance(resource_properties, list):
                                                                for resource_property in resource_properties:
                                                                    for default_key, default_value in default_gparams.items():
                                                                        if default_key not in resource_property:
                                                                            resource_property[default_key] = default_value
                                                                    resource_property["compiletime_identity"] = "module.%s" % key
                                                if resource not in new_resources:
                                                    new_resources[resource] = [resource_item]
                                                else:
                                                    new_resources[resource].append(resource_item)
                            else:
                                logger.error("module does not exist : %s ", value["source"])
                            
                if "module" in gen_template_json:
                    del gen_template_json["module"]

            if 'data' in template_json:
                data_resource = {}
                for data_item in template_json['data']:
                    for data_key, data_value in data_item.items():
                        processed_data, processed = self.process_resource(data_value, count=self.count)
                        if processed and isinstance(processed_data, dict):
                            for processed_key, processed_value in processed_data.items():
                                if isinstance(processed_value, dict):
                                    processed_value["compiletime_identity"] = "data.%s.%s" % (data_key, processed_key)
                                if isinstance(processed_value, list):
                                    for resource_property in processed_value:
                                        resource_property["compiletime_identity"] = "data.%s.%s" % (data_key, processed_key)
                            
                        self.gdata[data_key] = processed_data
                        data_resource[data_key] = processed_data
                gen_template_json['data'] = data_resource
            
            self.resource = {}
            resources = []
            if "resource" in new_resources and isinstance(new_resources["resource"], list):
                for resource in new_resources["resource"]:
                    for resource_name, properties in resource.items():
                        processed_resource, processed = self.process_resource(properties, count=self.count)
                        if not self.process_module:
                            if resource_name in self.resource:
                                self.resource[resource_name].append(processed_resource)
                            else:
                                self.resource[resource_name] = [processed_resource]
                        else:
                            self.resource[resource_name] = processed_resource

            if 'resource' in template_json:
                for res in template_json['resource']:
                    for resource_name, properties in res.items():
                        processed_resource, processed = self.process_resource(properties, count=self.count)
                        if not self.process_module:
                            if resource_name in self.resource:
                                self.resource[resource_name].append(processed_resource)
                            else:
                                self.resource[resource_name] = [processed_resource]
                        else:
                            if resource_name in self.resource and isinstance(self.resource[resource_name], dict) and isinstance(processed_resource, dict):
                                self.resource[resource_name].update(processed_resource)
                            else:    
                                self.resource[resource_name] = processed_resource

            if not self.process_module:
                for resource_name, processed_resource_list in self.resource.items():
                    for processed_resource in processed_resource_list:
                        if isinstance(processed_resource, dict):
                            for name, properties in processed_resource.items():
                                if isinstance(properties, list):
                                    for property_obj in properties:
                                        self.resource_types.append(resource_name.lower())
                                        if "compiletime_identity" not in property_obj:
                                            property_obj["compiletime_identity"] = "%s.%s" % (resource_name.lower(), name)
                                        resources.append({
                                            "type" : resource_name,
                                            "name" : name,
                                            "properties" : property_obj
                                        })    
                                else:
                                    self.resource_types.append(resource_name.lower())
                                    if "compiletime_identity" not in properties:
                                        properties["compiletime_identity"] = "%s.%s" % (resource_name.lower(), name)
                                    resources.append({
                                        "type" : resource_name,
                                        "name" : name,
                                        "properties" : properties
                                    })
                    gen_template_json['resources'] = resources

                    if 'resource' in gen_template_json:
                        del gen_template_json['resource']
            else:
                for parameter_json in parameter_jsons:
                    for file_type, variable_json in parameter_json.items():
                        if file_type == "tf" and "output" in variable_json:
                            for output in variable_json["output"]:
                                for key, value in output.items():
                                    if "value" in value:
                                        processed_data, _ = self.process_resource(value["value"], count=self.count)
                                        self.outputs[key] = processed_data
                gen_template_json['resource'] = self.resource
            
        return gen_template_json
    
    def check_numeric_value(self, resource):
        """ check that resource is numeric value or not and return the numeric value """
        try:
            resource = int(resource)
            return True, resource
        except ValueError:
            pass

        try:
            resource = float(resource)
            return True, resource
        except ValueError:
            pass

        return False, resource
    
    def check_json_or_list_value(self, resource, count=None):
        """ check that string resource is json or list type or not and return the list or json value """
        json_data = json_from_string(resource.replace("\'","\""))
        if json_data:
            resource, processed = self.process_resource(json_data, count=count)
            return True, resource
        
        json_data = json_from_string(re.sub(r"(?<!\\)\"", "\\\"", resource).replace("\'","\""))
        if json_data:
            resource, processed = self.process_resource(json_data, count=count)
            return True, resource
        
        try:
            if resource.startswith('[') and resource.endswith(']'):
                update_resource, processed = self.process_resource(resource[1:-1], count=count)
                list_data = ast.literal_eval("[" + str(update_resource) + "]")
                resource, processed = self.process_resource(list_data, count=count)
                return True, resource
        except:
            pass

        return False, resource
    
    def parse_string(self, resource):
        if resource.startswith('"') and resource.endswith('"'):
            resource = resource[1:-1]
        elif resource.startswith("'") and resource.endswith("'"):
            resource = resource[1:-1]
        return resource

    def split_parameters(self, value):
        try:
            value = "[%s]" % value
            value = value.replace("'", '"')
            parsed = hcl2.loads('split = %s \n' % value)
            
            params = []
            for split_str in parsed["split"]:
                if split_str != None:
                    split_str = str(split_str)
                    exmatch = re.search(r'^\${.*}$', split_str, re.I)
                    if exmatch:
                        match_values = re.search(r'(?<=\{).*(?=\})', split_str, re.I)
                        if match_values:
                            split_str = match_values.group(0)
                        else:
                            split_str = exmatch.group(0)[2:-1]
                
                params.append(split_str)
            return params 
        except Exception as e:
            logger.debug("Failed to split paramaters")
            logger.debug(value)
            return []

    def process_expression_parameters(self, param_str, count):
        
        groups = re.findall(r'([.a-zA-Z]+)[(].*[,].*[)]', param_str, re.I)
        if groups:
            for group in groups:
                updated_group, _ = self.process_resource(group, count)
                param_str.replace(group, updated_group)

        groups = re.findall(r'^[(].*[,].*[)]|.* ([(].*[)])', param_str, re.I)
        if groups:
            for group in groups:
                parameter_str = re.findall("(?<=\().*(?=\))", group)[0]
                updated_group = self.process_expression_parameters(parameter_str, count)
                param_str.replace(group, updated_group)

        return param_str

    # TODO: Fix this method
    # def process_expression_parameters(self, param_str, count):
        
    #     groups = re.findall(r'([.a-zA-Z]+[(].*[,].*[)])', param_str, re.I)
    #     if groups:
    #         for group in groups:
    #             updated_group, processed = self.process_resource(group, count)
    #             if processed:
    #                 param_str = param_str.replace(group, str(updated_group))
        
    #     groups = re.findall(r'^[(].*[,].*[)]|.* ([(].*[)])', param_str, re.I)
    #     if groups:
    #         for group in groups:
    #             parameter_str = re.findall("(?<=\().*(?=\))", group)[0]
    #             updated_group = self.process_expression_parameters(parameter_str, count)
    #             param_str = param_str.replace(group, str(updated_group))
        
    #     return param_str

    def process_resource(self, resource, count=None, nested_string_params={}):
        """ 
        process the resource json and return the resource with updated values
        """
        processed = True
        new_resource = ""
        if isinstance(resource, list):
            new_resource_list = [] 
            for value in resource:
                processed_resource, processed = self.process_resource(value, count=count)
                new_resource_list.append(processed_resource)
            new_resource = new_resource_list
        
        elif isinstance(resource, dict):
            new_resource = {}
            new_resource_list = []
            r_count = resource.get("count")
            if r_count:
                count_resource, processed = self.process_resource(r_count, count=count)
                if isinstance(count_resource, str):
                    result, res = self.check_numeric_value(count_resource)
                    if result:
                        count_resource = res
                    
                if isinstance(count_resource, int):
                    for i in range(count_resource):
                        new_resource_dict = {}
                        process_resource = copy.deepcopy(resource)
                        process_resource["count"] = i
                        del process_resource["count"]
                        self.count = i
                        for key, value in process_resource.items():
                            if key in self.skip_key_to_process:
                                new_resource_dict[key] = value
                                continue
                            elif key == "dynamic" and value and isinstance(value, list) and isinstance(value[0], dict):
                                processed_resource, processed = self.process_resource({ "dynamic" : value }, count=i)
                                if processed_resource and isinstance(processed_resource, dict):
                                    for res, val in processed_resource.items():
                                        new_resource_dict[res] = val
                            else:
                                processed_resource, processed = self.process_resource(value, count=i)
                                new_resource_dict[key] = processed_resource
                        new_resource_list.append(new_resource_dict)
                    self.count = None
                    new_resource = new_resource_list
                else:
                    for key, value in resource.items():
                        processed_resource, processed = self.process_resource(value, count=count)
                        new_resource[key] = processed_resource
            else:
                for key, values in resource.items():
                    if key in self.skip_key_to_process:
                        new_resource[key] = values
                        continue
                    # if key == "dynamic" and isinstance(value, dict):
                    if key == "dynamic" and values and isinstance(values, list) and isinstance(values[0], dict):
                        for value in values:
                        # value = value[0]
                            loop_values = []
                            for main_key, loop_content in value.items():
                                var = main_key
                                resource_properties = []
                                if isinstance(loop_content, dict):
                                    for loop_content_key, loop_content_value in loop_content.items():
                                        if loop_content_key == "for_each":
                                            loop_values, processed = self.process_resource(loop_content_value, count=count)
                                        if loop_content_key == "iterator":
                                            var, processed = self.process_resource(loop_content_value, count=count)
                                        # if loop_content_key == "content" and isinstance(loop_content_value, dict):
                                        if loop_content_key == "content" and loop_content_value and \
                                            isinstance(loop_content_value, list) and isinstance(loop_content_value[0], dict):
                                            loop_content_value = loop_content_value[0]
                                            if isinstance(loop_values, list):
                                                for loop_value in loop_values:
                                                    resource_property = {}
                                                    for content_key, content_value in loop_content_value.items():
                                                        if isinstance(content_value, str):
                                                            self.temp_params[var] = {
                                                                "value" : loop_value
                                                            }
                                                            # content_value = content_value.replace(var+".value",str(loop_value))
                                                        content_value, processed = self.process_resource(content_value, count=count)
                                                        resource_property[content_key] = content_value
                                                        self.temp_params = {}
                                                    resource_properties.append(resource_property)
                                            elif isinstance(loop_values, dict):
                                                for loop_key, loop_value in loop_values.items():
                                                    resource_property = {}
                                                    for content_key, content_value in loop_content_value.items():
                                                        if isinstance(content_value, str):
                                                            self.temp_params[var] = {
                                                                "value" : loop_value,
                                                                "key" : loop_key
                                                            }
                                                            # content_value = content_value.replace(var+".value",str(loop_value))
                                                            # content_value = content_value.replace(var+".key",str(loop_key))
                                                        content_value, processed = self.process_resource(content_value, count=count)
                                                        resource_property[content_key] = content_value
                                                        self.temp_params = {}
                                                    resource_properties.append(resource_property)
                                new_resource[main_key] = resource_properties
                    else:
                        processed_resource, processed = self.process_resource(values, count=count)
                        new_resource[key] = processed_resource
        
        elif isinstance(resource, str):
            # match the substrings for replace the value
            # pattern = re.compile(r'(var\..\w*)')
            # exmatch = re.findall(pattern, resource)
            parsed_string = self.parse_string(resource)

            if re.match(r"^\{.*\}$", parsed_string.strip().replace("\n", "")):
                json_data = json_from_string(re.sub(r"\n", "", parsed_string).replace("\\",""))
                if json_data:
                    parsed_string, processed = self.process_resource(json_data, count=count)
                    return parsed_string, True

            new_resource = copy.deepcopy(parsed_string)
            match_full = re.match(r'^\$\{([^}$]*)\}$', new_resource)
            if match_full:
                matched_str = new_resource[2:-1]
            else:
                expression_parameter = False
                for func in expression_list:
                    m = re.match(func['expression'], new_resource)
                    if m:
                        expression_parameter = True
                    
                all_exmatch = re.findall(r'\${([^}$]*)}', new_resource)
                if all_exmatch and not expression_parameter:
                    for exmatch in all_exmatch:
                        processed_param, processed = self.process_resource("${" + copy.deepcopy(exmatch).strip() + "}", count=count)
                        parsed_string = parsed_string.replace("${" + exmatch + "}", str(processed_param))
                    new_resource = copy.deepcopy(parsed_string)
                    matched_str = parsed_string
                else:
                    matched_str = parsed_string

            exmatch = re.search(r'\${([^}]*)}', matched_str, re.I)
            if exmatch:
                match_values = re.search(r'(?<=\{).*(?=\})', matched_str, re.I)
                if match_values:
                    matched_str = match_values.group(0)
                else:
                    matched_str = exmatch.group(0)[2:-1]
            matched_str = matched_str.strip()
            matched_str = self.parse_string(matched_str)

            # if not re.match(r'^([.a-zA-Z]+[(].*[,].*[)])$', matched_str) and \
            #     not re.match(r'^[(].*[,].*[)]|.* ([(].*[)])', matched_str):
            #     matched_str = self.process_expression_parameters(matched_str, count)
                
            result, res = self.check_numeric_value(matched_str)
            if result:
                new_resource = res
                return new_resource, processed

            result, res = self.check_json_or_list_value(matched_str, count=count)
            if result:
                new_resource = res
                return new_resource, processed

            for func in default_functions:
                m = re.match(func['expression'], matched_str)
                if not m:
                    continue

                parameter_str = re.findall("(?<=\().*(?=\))", m.group(0))[0]
                parameters = []
                process = True

                found_parameters = self.split_parameters(parameter_str.strip())

                for param in found_parameters:
                # for param in re.findall("(?:[^,()]|\((?:[^()]|\((?:[^()]|\([^()]*\))*\))*\))+", parameter_str.strip()):
                # for param in parameter_str.strip().split(","):
                    param_exmatch = re.search(r'(^[\'|\"]\${([^}]*)}[\'|\"]$)|(^\${([^}]*)}$)', param.strip(), re.I)
                    if param_exmatch:
                        match_values = re.search(r'(?<=\{).*(?=\})', param.strip(), re.I)
                        if match_values:
                            param = match_values.group(0)
                    if param in nested_string_params:
                        param = str(nested_string_params[param])
                
                    processed_param, process_status = self.process_resource("${" + param.strip() + "}", count)
                    processed = processed and process_status
                    if (isinstance(processed_param, str) and re.search(r'\${([^}]*)}', processed_param, re.I)):
                        process = False
                        parameters.append(param.strip())
                    # Check only for None return
                    elif processed_param == None and re.match(".*(\.\*\.).*", param.strip()):
                        process = False
                        parameters.append(param.strip())
                    else:
                        parameters.append(processed_param)

                args = inspect.getargspec(func['method']).args
                varargs = inspect.getargspec(func['method']).varargs
                keywords = inspect.getargspec(func['method']).keywords
                defaults = inspect.getargspec(func['method']).defaults

                if process and ((len(args) == len(parameters)) or \
                    (defaults and len(parameters) <= len(args) and len(parameters) >= (len(args) - len(defaults))) or \
                    varargs or keywords):
                    try:
                        new_resource = func['method'](*parameters)
                    except Exception as e:
                        logger.error("Failed to process %s : %s", new_resource, str(e))
                else:
                    parameters = [str(ele) for ele in parameters]
                    new_resource = func['method'].__name__ + "(" + ",".join(parameters) + ")"
                break 
            else:
                for func in expression_list:
                    m = re.match(func['expression'], matched_str)
                    if not m:
                        continue

                    parameter_str = m.group(0)
                    parameter_str = self.process_expression_parameters(parameter_str, count)
                    
                    string_params = {}
                    groups = re.findall(r'\".*?\"', parameter_str, re.I)
                    if groups:
                        i = 0
                        for group in groups:
                            updated_group, _ = self.process_resource(groups[i], count)
                            string_params["group%s" % (str(i))] = updated_group
                            parameter_str = parameter_str.replace(group, "group%s" % (str(i)))
                            i+=1
                    
                    # params = re.findall(r"[a-zA-Z0-9.()\[\]_*\"]+|(?:(?![a-zA-Z0-9.()\[\]_*\"]).)+", parameter_str)
                    params = re.findall(r"[a-zA-Z0-9.()\[\],-_*\"\'\{\$\}]+|(?:(?![ a-zA-Z0-9.()\[\]_*\"\'\{\$\}]).)+", parameter_str)

                    if params and len(params) > 1:
                        new_parameter_list = []
                        process_function = True
                        for param in params:
                            if param in string_params:
                                param = str(string_params[param])
                            if param in nested_string_params:
                                param = str(nested_string_params[param])

                            processed_param, processed = self.process_resource("${" + param.strip() + "}", count=count, nested_string_params=string_params)
                            if isinstance(processed_param, str) and (re.findall(r".*\(.*\)", processed_param) or re.findall(r"\${([^}]*)}", processed_param)):
                                process_function = False # parameter processing failed
                                new_parameter_list.append("\"" + param + "\"")
                            elif isinstance(processed_param, str) and re.findall(r"[a-zA-Z]", processed_param) and processed_param not in self.replace_value_str:
                                new_parameter_list.append("\"" + processed_param + "\"")
                            elif isinstance(processed_param, str) and len(processed_param) == 0:
                                new_parameter_list.append('""')
                            elif processed_param is None and len(param.strip().split(".")) > 1 and not processed:
                                process_function = False # variable value not set in vars.tf file
                                new_parameter_list.append("\"" + param + "\"")
                            else:
                                new_parameter_list.append(str(processed_param))

                        if process_function:
                            new_resource = func['method'](" ".join(new_parameter_list))
                        else:
                            new_resource = matched_str
                else:
                    splited_list = matched_str.split(".") 
                    if len(splited_list) > 1:
                        if splited_list[0] in self.schema_filter:
                            result, new_value = self.schema_filter[splited_list[0]](".".join(splited_list[1:]))
                            if result:
                                new_resource = new_value
                            else:
                                processed = False
                        else:
                            if matched_str == "count.index" and count is not None:
                                new_resource = count
                            else:
                                result, new_value = self.schema_filter["other"](".".join(splited_list))
                                if result:
                                    new_resource = new_value
                    else:
                        new_resource = matched_str
        else:
            new_resource = resource
        
        if isinstance(new_resource, str) and new_resource.strip() in self.replace_values:
            new_resource = self.replace_values[new_resource.strip()]

        return new_resource, processed



import os
import re
import tempfile
import urllib.parse
import stat
from subprocess import Popen, PIPE
from processor.connector.git_connector.git_functions import GithubFunctions, check_clone_repos, set_clone_repo
from processor.helper.file.file_utils import exists_file, exists_dir
from processor.helper.json.json_utils import get_field_value, get_field_value_with_default
from processor.logging.log_handler import getlogger
from processor.helper.config.rundata_utils import get_from_currentdata
from processor.connector.vault import get_vault_data

logger = getlogger()

def create_ssh_file_vault_data(ssh_dir, ssh_key_file_data, ssh_fname):
    flags = os.O_WRONLY | os.O_CREAT | os.O_EXCL  # Refer to "man 2 open".
    mode = stat.S_IRUSR | stat.S_IWUSR
    umask = 0o777 ^ mode
    umask_original = os.umask(umask)
    ssh_key_file = '%s/%s' % (ssh_dir, ssh_fname)
    try:
        fdesc = os.open(ssh_key_file, flags, mode)
        with os.fdopen(fdesc, 'w') as fout:
            for l in  ssh_key_file_data.split('\\n'):
                fout.write(l + '\n')
    finally:
        os.umask(umask_original)
    return ssh_key_file

def run_subprocess_cmd(cmd, ignoreerror=False, maskoutput=False, outputmask="Error output is masked"):
    """ Run a sub-process command"""
    result = ''
    errresult = None
    if cmd:
        if isinstance(cmd, list):
            cmd = ' '.join(cmd)
        myprocess = Popen(cmd, shell=True, stdout=PIPE, stderr=PIPE, stdin=PIPE)
        out, err = myprocess.communicate()
        result = out.rstrip()
        errresult = err.rstrip() if err else None
        if isinstance(result, bytes):
            result = result.decode()
        if errresult and isinstance(errresult, bytes):
            errresult = errresult.decode()
    return errresult, result

def create_ssh_config(ssh_dir, ssh_key_file, ssh_user):
    ssh_config = '%s/config' % ssh_dir
    if exists_file(ssh_config):
        logger.error("Git config: %s already exists, cannot modify it!")
        return None
    with open(ssh_config, 'w') as f:
        f.write('Host *\n')
        # f.write('Host %s\n' % ssh_host)
        # f.write('   HostName %s\n' % ssh_host)
        f.write('   User %s\n' % ssh_user)
        f.write('   IdentityFile %s\n\n' % ssh_key_file)
        # f.write('Host *\n')
        f.write('   IdentitiesOnly yes\n')
        f.write('   ServerAliveInterval 100\n')
    return ssh_config

def get_pwd_from_vault(password_key):
    """ Return the git password from vault """
    password = get_vault_data(password_key)
    if not password:
        logger.info("Password does not set in the vault")
    return password

def get_git_pwd(key='GIT_PWD'):
    """ Return the git password for https connection"""
    git_pwd = get_from_currentdata('GIT_PWD')
    if not git_pwd:
        git_pwd = os.getenv(key, None)
        if git_pwd:
            logger.info("Git password from envirnment: %s", '*' * len(git_pwd))
    return git_pwd

def git_clone_dir(connector, giturl=None, branch=None, clone_specific_branch=True):
    global CLONE_REPOS
    clonedir = None
    repopath = tempfile.mkdtemp()
    subdir = False
    if connector and isinstance(connector, dict):
        if not giturl:
            giturl = get_field_value(connector, 'gitProvider')
            if not giturl:
                logger.error("Git connector does not have valid git provider URL")
                return repopath, clonedir
        
        if not branch and clone_specific_branch:
            # Removed command line option for branch
            branch = get_field_value_with_default(connector, 'branchName', 'master')
            # branch = get_from_currentdata('branch')
            # if not branch:
            #     branch = get_field_value_with_default(connector, 'branchName', 'master')

        isprivate = get_field_value(connector, 'private')
        isprivate = True if isprivate is None or not isinstance(isprivate, bool) else isprivate
        # logger.info("Repopath: %s", repopath)
        logger.info("\t\t\tRepopath: %s", repopath)
        http_match = re.match(r'^http(s)?://', giturl, re.I)
        if http_match:
            logger.info("\t\t\tHttp (private:%s) giturl: %s", "YES" if isprivate else "NO", giturl)

            accessToken = get_field_value(connector, 'httpsAccessToken')
            username = get_field_value(connector, 'httpsUser')
            if accessToken:
                logger.info("AccessToken: %s" % accessToken)
                pwd = get_field_value(connector, 'httpsPassword')
                isremote = get_from_currentdata('remote')
                if isremote:
                    gittoken = get_from_currentdata('gittoken')
                    pwd = gittoken if gittoken else None
                pwd = pwd if pwd else get_git_pwd(key=accessToken)
                if not pwd:
                    pwd = get_pwd_from_vault(accessToken)
                    if pwd:
                        logger.info("Git access token from vault: %s", '*' * len(pwd))
                if pwd:
                    gh = GithubFunctions()
                    gh.set_base_url(giturl)
                    gh.set_access_token(pwd)
                    _ = gh.populate_user()
                    rpo = gh.clone_repo(giturl, repopath, branch)
                    if rpo:
                        logger.info('Successfully cloned in %s dir' % repopath)
                        checkdir = '%s/tmpclone' % repopath if subdir else repopath
                        clonedir = checkdir if exists_dir('%s/.git' % checkdir) else None
                        if not exists_dir(clonedir):
                            logger.error("No valid data provided for connect to git : %s", giturl)
                        return repopath, clonedir
                    elif isprivate:
                        logger.error("Please provide password for connect to git repository.")
                        return repopath, clonedir
                    else:
                        git_cmd = 'git clone %s %s' % (giturl, repopath)
            elif username:
                pwd = get_field_value(connector, 'httpsPassword')
                schema = giturl[:http_match.span()[-1]]
                other_part = giturl[http_match.span()[-1]:]
                # pwd = pwd if (pwd and not json_source()) else (get_git_pwd() if not json_source() else get_pwd_from_vault(pwd))
                pwd = pwd if pwd else get_git_pwd(key=username)

                # populate the password from vault
                if not pwd:
                    pwd = get_pwd_from_vault(username)
                    if pwd:
                        logger.info("Git password from vault: %s", '*' * len(pwd))
                if pwd:
                    git_cmd = 'git clone --depth 1 %s%s:%s@%s %s' % (schema, urllib.parse.quote_plus(username),
                                                        urllib.parse.quote_plus(pwd), other_part, repopath)
                elif isprivate:
                    logger.error("Please provide password for connect to git repository.")
                    return repopath, clonedir
                else:
                    git_cmd = 'git clone --depth 1 %s%s:%s@%s %s' % (schema, urllib.parse.quote_plus(username), "",
                                                     other_part, repopath)
            else:
                git_cmd = 'git clone --depth 1 %s %s' % (giturl, repopath)
        else:
            logger.info("SSH (private:%s) giturl: %s, Repopath: %s", "YES" if isprivate else "NO",
                        giturl, repopath)
            if isprivate:
                ssh_key_file = get_field_value(connector, 'sshKeyfile')
                ssh_key_name = get_field_value(connector, 'sshKeyName')
                ssh_key_file_data = None
                if ssh_key_file:
                    if not exists_file(ssh_key_file):
                        logger.error("Git connector points to a non-existent ssh keyfile!")
                        return repopath, clonedir
                elif ssh_key_name:
                    ssh_key_file_data = get_vault_data(ssh_key_name)
                    if not ssh_key_file_data:
                        logger.info('Git connector points to a non-existent ssh keyName in the vault!')
                        return repopath, clonedir
                ssh_host = get_field_value(connector, 'sshHost')
                ssh_user = get_field_value_with_default(connector, 'sshUser', 'git')
                if not ssh_host:
                    logger.error("SSH host not set, could be like github.com, gitlab.com, 192.168.1.45 etc")
                    return repopath, clonedir
                ssh_dir = '%s/.ssh' % repopath
                if exists_dir(ssh_dir):
                    logger.error("Git ssh dir: %s already exists, cannot recreate it!", ssh_dir)
                    return repopath, clonedir
                os.mkdir('%s/.ssh' % repopath, 0o700)
                if not ssh_key_file and ssh_key_name and ssh_key_file_data:
                    ssh_key_file = create_ssh_file_vault_data(ssh_dir, ssh_key_file_data, ssh_key_name)
                    if not ssh_key_file:
                        logger.info('Git connector points to a non-existent ssh keyName in the vault!')
                        return repopath, clonedir
                ssh_cfg = create_ssh_config(ssh_dir, ssh_key_file, ssh_user)
                if not ssh_cfg:
                    logger.error("Creation of Git ssh config in dir: %s failed!", ssh_dir)
                    return repopath, clonedir
                git_ssh_cmd = 'ssh -o "StrictHostKeyChecking=no" -F %s' % ssh_cfg
                git_cmd = 'git clone %s %s/tmpclone' % (giturl, repopath)
                subdir = True
            else:
                git_ssh_cmd = 'ssh -o "StrictHostKeyChecking=no"'
                git_cmd = 'git clone %s %s' % (giturl, repopath)
            os.environ['GIT_SSH_COMMAND'] = git_ssh_cmd
        
        if branch:
            git_cmd = '%s --branch %s' % (git_cmd, branch)
        
        if git_cmd:
            re_path, cln_dir = check_clone_repos(git_cmd=git_cmd.replace(repopath, ""))
            if re_path and cln_dir:
                return re_path, cln_dir
            error_result, result = run_subprocess_cmd(git_cmd)
            checkdir = '%s/tmpclone' % repopath if subdir else repopath
            clonedir = checkdir if exists_dir('%s/.git' % checkdir) else None
            if not exists_dir(clonedir):
                logger.error("No valid data provided for connect to git : %s", error_result)
            else:
                set_clone_repo(git_cmd.replace(repopath, ""), repopath, clonedir)
        if 'GIT_SSH_COMMAND' in os.environ:
            os.environ.pop('GIT_SSH_COMMAND')
    return repopath, clonedir



"""
Performs all in built string functions which are supported by terraform processor
"""
from processor.logging.log_handler import getlogger
import decimal

logger = getlogger()

def chomp(str_value):
    """ removes newline characters at the end of a string. """
    return str_value.rstrip()

# def format(str_value):
#     """ removes newline characters at the end of a string. """
#     return str_value.rstrip()
    
def join(concat_ele, ele_list):
    """ concat list items and returns a string """
    return concat_ele.join(ele_list)

def lower(str_value):
    """ convert string to lower case """
    return str_value.lower()

def replace(str_value, substring, replacement):
    """ searches for substring and replace the value of substring """
    return str_value.replace(substring , replacement)

def split(separator, str_value):
    """ split the string by given separator and returns the list """
    return str_value.split(separator) 

def trim(str_value, trim_string):
    """ trim the characters from the given string """
    return str_value.strip(trim_string)

def trimprefix(str_value, trim_string):
    """ trim the characters from the given string """
    if str_value.startswith(trim_string):
        str_value = str_value[len(trim_string):]
        return trimprefix(str_value, trim_string)
    return str_value

def trimsuffix(str_value, trim_string):
    """ trim the characters from the given string """
    if str_value.endswith(trim_string):
        str_value = str_value[:-(len(trim_string))]
        return trimsuffix(str_value, trim_string)
    return str_value

def trimspace(str_value):
    """ trim the space from the given string """
    return str_value.strip()
    
def upper(str_value):
    """ convert string to upper case """
    return str_value.upper()

def strrev(str_value):
    """ reverse the characters of given string """
    return str_value[::-1]

def substr(str_value, offset, length):
    """ return the subsctring of given string """
    return str_value[offset:length]

def title(str_value):
    """ converts the first character of each word of given string to uppercase. """
    return str_value.title()
    
def format(spec, *values):
    """ format the string """
    return spec % values
    